# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from networking_cisco._i18n import _

from neutronclient.common import extension


class NetworkProfile(extension.NeutronClientExtension):
    resource = 'network_profile'
    resource_plural = '%ss' % resource
    object_path = '/%s' % resource_plural
    resource_path = '/%s/%%s' % resource_plural
    versions = ['2.0']
    segment_types = ['vlan', 'overlay']
    segment_sub_types = ['native', 'enhanced']
    allow_names = True


class NetworkProfileList(extension.ClientExtensionList, NetworkProfile):
    """List network profiles that belong to a given tenant."""

    shell_command = 'cisco-network-profile-list'
    list_columns = ['id', 'name', 'segment_type', 'sub_type', 'segment_range',
                    'physical_network', 'multicast_ip_index',
                    'multicast_ip_range']
    pagination_support = True
    sorting_support = True


class NetworkProfileShow(extension.ClientExtensionShow, NetworkProfile):
    """Show information of a given network profile."""

    shell_command = 'cisco-network-profile-show'


class NetworkProfileCreate(extension.ClientExtensionCreate, NetworkProfile):
    """Create a network profile."""

    shell_command = 'cisco-network-profile-create'

    def add_known_arguments(self, parser):
        parser.add_argument('name',
                            help=_('Name for network profile.'))
        parser.add_argument('segment_type',
                            choices=self.segment_types,
                            help='Segment type.')
        parser.add_argument('--sub_type',
                            choices=self.segment_sub_types,
                            help=_('Sub-type for the segment. Available '
                                   'sub-types for overlay segments: '
                                   'native, enhanced.'))
        parser.add_argument('--segment_range',
                            help=_('Range for the segment.'))
        parser.add_argument('--physical_network',
                            help=_('Name for the physical network.'))
        parser.add_argument('--multicast_ip_range',
                            help=_('Multicast IPv4 range.'))
        parser.add_argument("--add-tenant",
                            action='append', dest='add_tenants',
                            help=_("Add tenant to the network profile. "
                                   "You can repeat this option."))

    def args2body(self, parsed_args):
        body = {self.resource: {'name': parsed_args.name}}
        if parsed_args.segment_type:
            body[self.resource].update({'segment_type':
                                        parsed_args.segment_type})
        if parsed_args.sub_type:
            body[self.resource].update({'sub_type':
                                        parsed_args.sub_type})
        if parsed_args.segment_range:
            body[self.resource].update({'segment_range':
                                        parsed_args.segment_range})
        if parsed_args.physical_network:
            body[self.resource].update({'physical_network':
                                        parsed_args.physical_network})
        if parsed_args.multicast_ip_range:
            body[self.resource].update({'multicast_ip_range':
                                        parsed_args.multicast_ip_range})
        if parsed_args.add_tenants:
            body[self.resource].update({'add_tenants':
                                        parsed_args.add_tenants})
        return body


class NetworkProfileDelete(extension.ClientExtensionDelete, NetworkProfile):
    """Delete a given network profile."""

    shell_command = 'cisco-network-profile-delete'



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function

import argparse
import six

from networking_cisco._i18n import _

from neutronclient.common import extension
from neutronclient.neutron import v2_0 as neutronV20

DEVICE = 'hosting_device'
HOSTING_DEVICE_CONFIG = '/get_hosting_device_config'


def _add_updatable_args(parser):
    parser.add_argument(
        '--credentials-id',
        help=_('Id of credentials used by this hosting device.'))
    parser.add_argument(
        '--credentials_id',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--name',
        help=_('Name of this hosting device'))
    parser.add_argument(
        '--description',
        help=_('Description of this hosting device.'))
    parser.add_argument(
        '--device-id',
        help=_('Manufacturer id of hosting device.'))
    parser.add_argument(
        '--device_id',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--management-ip-address',
        help=_('IP address used for management of hosting device.'))
    parser.add_argument(
        '--management_ip_address',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--protocol-port',
        help=_('Protocol port used for management of hosting device.'))
    parser.add_argument(
        '--protocol_port',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--admin-state-down',
        dest='admin_state_up',
        action='store_false',
        help=_('Set hosting device administratively down.'),
        default=argparse.SUPPRESS,)
    parser.add_argument(
        '--admin_state_down',
        dest='admin_state_up',
        action='store_false',
        help=argparse.SUPPRESS,
        default=argparse.SUPPRESS,)
    parser.add_argument(
        '--tenant-bound',
        help=_('Tenant allowed place service instances in the hosting '
               'device.'),
        default=argparse.SUPPRESS)
    parser.add_argument(
        '--tenant_bound',
        help=argparse.SUPPRESS,
        default=argparse.SUPPRESS)
    parser.add_argument(
        '--auto-delete',
        dest='auto_delete',
        action='store_true',
        help=_('Make hosting device subject to automated life cycle '
               'management.'),
        default=argparse.SUPPRESS)
    parser.add_argument(
        '--auto_delete',
        dest='auto_delete',
        action='store_true',
        help=argparse.SUPPRESS,
        default=argparse.SUPPRESS)


def _updatable_args2body(parsed_args, body):
    neutronV20.update_dict(parsed_args, body[DEVICE], [
        'credentials_id', 'name', 'description', 'device_id',
        'management_ip_address', 'protocol_port', 'admin_state_up',
        'auto_delete'])
    # handle tenant_bound separately as we want to allow it to be set to None
    if hasattr(parsed_args, 'tenant_bound'):
        if (isinstance(parsed_args.tenant_bound, six.string_types) and
                parsed_args.tenant_bound.lower() == 'none'):
            parsed_args.tenant_bound = None
        body[DEVICE].update({'tenant_bound': parsed_args.tenant_bound})


class HostingDevice(extension.NeutronClientExtension):
    resource = DEVICE
    resource_plural = '%ss' % resource
    object_path = '/dev_mgr/%s' % resource_plural
    resource_path = '/dev_mgr/%s/%%s' % resource_plural
    versions = ['2.0']
    allow_names = True


class HostingDeviceList(extension.ClientExtensionList, HostingDevice):
    """List hosting devices that belong to a given tenant."""

    shell_command = 'cisco-hosting-device-list'
    list_columns = ['id', 'name', 'template_id', 'admin_state_up', 'status']
    pagination_support = True
    sorting_support = True


class HostingDeviceShow(extension.ClientExtensionShow, HostingDevice):
    """Show information of a given hosting device."""

    shell_command = 'cisco-hosting-device-show'


class HostingDeviceCreate(extension.ClientExtensionCreate, HostingDevice):
    """Create a hosting device for a given tenant."""

    shell_command = 'cisco-hosting-device-create'

    def add_known_arguments(self, parser):
        _add_updatable_args(parser)
        parser.add_argument(
            '--id',
            help=_('Id for this hosting device.'))
        parser.add_argument(
            '--management-port',
            help=_('Neutron port used for management of hosting device.'))
        parser.add_argument(
            '--management_port',
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--cfg-agent-id',
            help=_('Config agent to handle the hosting device.'))
        parser.add_argument(
            '--cfg_agent_id',
            help=argparse.SUPPRESS)

        parser.add_argument(
            'name', metavar='NAME',
            help=_('Name of hosting device to create.'))
        parser.add_argument(
            'template_id', metavar='TEMPLATE',
            help=_('Hosting device template template to associate '
                   'hosting device with.'))

    def args2body(self, parsed_args):
        client = self.get_client()
        _template_id = neutronV20.find_resourceid_by_name_or_id(
            client, 'hosting_device_template', parsed_args.template_id)
        body = {self.resource: {
            'name': parsed_args.name, 'template_id': _template_id,
            'admin_state_up': True}}
        _updatable_args2body(parsed_args, body)
        neutronV20.update_dict(parsed_args, body[self.resource],
                               ['id', 'tenant_id', 'management_port_id',
                                'cfg_agent_id'])
        if (hasattr(parsed_args, 'management_port') and
                parsed_args.management_port):
            _port_id = neutronV20.find_resourceid_by_name_or_id(
                client, 'port', parsed_args.management_port)
            body[self.resource]['management_port_id'] = _port_id
        return body


class HostingDeviceDelete(extension.ClientExtensionDelete, HostingDevice):
    """Delete a given hosting device."""

    shell_command = 'cisco-hosting-device-delete'


class HostingDeviceUpdate(extension.ClientExtensionUpdate, HostingDevice):
    """Update hosting device's information."""

    shell_command = 'cisco-hosting-device-update'

    def add_known_arguments(self, parser):
        # adding admin_state_up here so that it is available for update only
        # as it is True by default and not meaningful in the create operation
        parser.add_argument(
            '--admin-state-up',
            dest='admin_state_up',
            action='store_true',
            help=_('Set hosting device administratively up.'),
            default=argparse.SUPPRESS)
        parser.add_argument(
            '--admin_state_up',
            dest='admin_state_up',
            action='store_true',
            help=argparse.SUPPRESS,
            default=argparse.SUPPRESS)
        # adding no_auto_delete here so that it is available for update only
        # as auto_delete is False by default and not meaningful in the create
        # operation
        parser.add_argument(
            '--no-auto-delete',
            dest='auto_delete',
            action='store_false',
            help=_('Exempt hosting device from automated life cycle '
                   'management.'),
            default=argparse.SUPPRESS)
        parser.add_argument(
            '--no_auto_delete',
            dest='auto_delete',
            action='store_false',
            help=argparse.SUPPRESS,
            default=argparse.SUPPRESS)
        _add_updatable_args(parser)

    def args2body(self, parsed_args):
        body = {self.resource: {}}
        _updatable_args2body(parsed_args, body)
        return body


class HostingDeviceGetConfig(extension.ClientExtensionShow, HostingDevice):
    """Fetch running of a given hosting device."""

    shell_command = 'cisco-hosting-device-get-config'

    def run(self, parsed_args):
        data = self.execute(parsed_args)
        # just do raw text output
        if data:
            self.app.stdout.write(str(data) + '\n')
        return 0

    def execute(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        neutron_client = self.get_client()
        neutron_client.format = parsed_args.request_format
        _id_hd = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'hosting_device', parsed_args.id)
        return self.get_hosting_device_config(neutron_client, _id_hd)

    def get_hosting_device_config(self, client, hosting_device_id):
        """Get config of hosting_device."""
        return client.get((self.resource_path + HOSTING_DEVICE_CONFIG) %
                          hosting_device_id)



# Copyright (c) 2013-2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import mock
import testtools

from networking_cisco.ml2_drivers.nexus import exceptions
from networking_cisco.ml2_drivers.nexus import nexus_db_v2
from oslo_db import exception as db_exc

from neutron.tests.unit import testlib_api


class TestCiscoNexusDb(testlib_api.SqlTestCase):

    """Unit tests for Cisco mechanism driver's Nexus port binding database."""

    NpbObj = collections.namedtuple('NpbObj', 'port vlan vni switch instance '
                 'channel_group is_native_vlan')

    def _npb_test_obj(self, pnum, vnum, vni=0, switch='10.9.8.7',
                      instance=None, channel_group=0, is_native_vlan=False):
        """Creates a Nexus port binding test object from a pair of numbers."""
        if pnum is 'router':
            port = pnum
        else:
            port = '1/%s' % pnum
        if instance is None:
            instance = 'instance_%s_%s' % (pnum, vnum)
        return self.NpbObj(port, vnum, vni, switch, instance,
                           channel_group, is_native_vlan=False)

    def _assert_bindings_match(self, npb, npb_obj):
        """Asserts that a port binding matches a port binding test obj."""
        self.assertEqual(npb.port_id, npb_obj.port)
        self.assertEqual(npb.vlan_id, npb_obj.vlan)
        self.assertEqual(npb.switch_ip, npb_obj.switch)
        self.assertEqual(npb.instance_id, npb_obj.instance)

    def _add_binding_to_db(self, npb):
        """Adds a port binding to the Nexus database."""
        return nexus_db_v2.add_nexusport_binding(
            npb.port, npb.vlan, npb.vni, npb.switch, npb.instance,
            npb.channel_group, npb.is_native_vlan)

    def _add_bindings_to_db(self, npbs):
        """Adds a list of port bindings to the Nexus database."""
        for npb in npbs:
            nexus_db_v2.add_nexusport_binding(
                npb.port, npb.vlan, npb.vni, npb.switch, npb.instance,
                npb.channel_group, npb.is_native_vlan)

    def _remove_binding_from_db(self, npb):
        """Removes a port binding from the Nexus database."""
        return nexus_db_v2.remove_nexusport_binding(
            npb.port, npb.vlan, npb.vni, npb.switch, npb.instance)

    def _get_nexusport_binding(self, npb):
        """Gets a port binding based on port, vlan, switch, and instance."""
        return nexus_db_v2.get_nexusport_binding(
            npb.port, npb.vlan, npb.switch, npb.instance)

    def _get_nexusvlan_binding(self, npb):
        """Gets port bindings based on vlan and switch."""
        return nexus_db_v2.get_nexusvlan_binding(npb.vlan, npb.switch)

    def _get_nexusvm_binding(self, npb):
        """Gets port binding based on vlan and instance."""
        return nexus_db_v2.get_nexusvm_bindings(npb.vlan, npb.instance)[0]

    def _get_port_vlan_switch_binding(self, npb):
        """Gets port bindings based on port, vlan, and switch."""
        return nexus_db_v2.get_port_vlan_switch_binding(
            npb.port, npb.vlan, npb.switch)

    def _get_port_switch_bindings(self, npb):
        """Get port bindings based on port and switch."""
        return nexus_db_v2.get_port_switch_bindings(npb.port, npb.switch)

    def test_nexusportbinding_add_remove(self):
        """Tests add and removal of port bindings from the Nexus database."""
        npb11 = self._npb_test_obj(10, 100)
        npb = self._add_binding_to_db(npb11)
        self._assert_bindings_match(npb, npb11)
        npb = self._remove_binding_from_db(npb11)
        self.assertEqual(len(npb), 1)
        self._assert_bindings_match(npb[0], npb11)
        with testtools.ExpectedException(exceptions.NexusPortBindingNotFound):
            self._remove_binding_from_db(npb11)

    def test_nexusportbinding_get(self):
        """Tests get of specific port bindings from the database."""
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100)
        npb22 = self._npb_test_obj(20, 200)
        self._add_bindings_to_db([npb11, npb21, npb22])

        npb = self._get_nexusport_binding(npb11)
        self.assertEqual(len(npb), 1)
        self._assert_bindings_match(npb[0], npb11)
        npb = self._get_nexusport_binding(npb21)
        self.assertEqual(len(npb), 1)
        self._assert_bindings_match(npb[0], npb21)
        npb = self._get_nexusport_binding(npb22)
        self.assertEqual(len(npb), 1)
        self._assert_bindings_match(npb[0], npb22)

        with testtools.ExpectedException(exceptions.NexusPortBindingNotFound):
            nexus_db_v2.get_nexusport_binding(
                npb21.port, npb21.vlan, npb21.switch, "dummyInstance")

    def test_nexusvlanbinding_get(self):
        """Test get of port bindings based on vlan and switch."""
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100)
        npb22 = self._npb_test_obj(20, 200)
        self._add_bindings_to_db([npb11, npb21, npb22])

        npb_all_v100 = self._get_nexusvlan_binding(npb11)
        self.assertEqual(len(npb_all_v100), 2)
        npb_v200 = self._get_nexusvlan_binding(npb22)
        self.assertEqual(len(npb_v200), 1)
        self._assert_bindings_match(npb_v200[0], npb22)

        with testtools.ExpectedException(exceptions.NexusPortBindingNotFound):
            nexus_db_v2.get_nexusvlan_binding(npb21.vlan, "dummySwitch")

    def test_nexusvmbinding_get(self):
        """Test get of port bindings based on vlan and instance."""
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100)
        npb22 = self._npb_test_obj(20, 200)
        self._add_bindings_to_db([npb11, npb21, npb22])

        npb = self._get_nexusvm_binding(npb21)
        self._assert_bindings_match(npb, npb21)
        npb = self._get_nexusvm_binding(npb22)
        self._assert_bindings_match(npb, npb22)

        with testtools.ExpectedException(exceptions.NexusPortBindingNotFound):
            nexus_db_v2.get_nexusvm_bindings(npb21.vlan, "dummyInstance")[0]

    def test_nexusportvlanswitchbinding_get(self):
        """Tests get of port bindings based on port, vlan, and switch."""
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100)
        self._add_bindings_to_db([npb11, npb21])

        npb = self._get_port_vlan_switch_binding(npb11)
        self.assertEqual(len(npb), 1)
        self._assert_bindings_match(npb[0], npb11)

        with testtools.ExpectedException(exceptions.NexusPortBindingNotFound):
            nexus_db_v2.get_port_vlan_switch_binding(
                npb21.port, npb21.vlan, "dummySwitch")

    def test_nexusportswitchbinding_get(self):
        """Tests get of port bindings based on port and switch."""
        npb11 = self._npb_test_obj(10, 100)
        npb21 = self._npb_test_obj(20, 100, switch='2.2.2.2')
        npb22 = self._npb_test_obj(20, 200, switch='2.2.2.2')
        self._add_bindings_to_db([npb11, npb21, npb22])

        npb = self._get_port_switch_bindings(npb11)
        self.assertEqual(len(npb), 1)
        self._assert_bindings_match(npb[0], npb11)
        npb_all_p20 = self._get_port_switch_bindings(npb21)
        self.assertEqual(len(npb_all_p20), 2)

        npb = nexus_db_v2.get_port_switch_bindings(npb21.port, "dummySwitch")
        self.assertIsNone(npb)

    def test_nexusbinding_update(self):
        """Tests update of vlan IDs for port bindings."""
        npb11 = self._npb_test_obj(10, 100, switch='1.1.1.1', instance='test')
        npb21 = self._npb_test_obj(20, 100, switch='1.1.1.1', instance='test')
        self._add_bindings_to_db([npb11, npb21])

        npb_all_v100 = nexus_db_v2.get_nexusvlan_binding(100, '1.1.1.1')
        self.assertEqual(len(npb_all_v100), 2)

        npb22 = self._npb_test_obj(20, 200, switch='1.1.1.1', instance='test')
        npb = nexus_db_v2.update_nexusport_binding(npb21.port, 200)
        self._assert_bindings_match(npb, npb22)

        npb_all_v100 = nexus_db_v2.get_nexusvlan_binding(100, '1.1.1.1')
        self.assertEqual(len(npb_all_v100), 1)
        self._assert_bindings_match(npb_all_v100[0], npb11)

        npb = nexus_db_v2.update_nexusport_binding(npb21.port, 0)
        self.assertIsNone(npb)

        npb33 = self._npb_test_obj(30, 300, switch='1.1.1.1', instance='test')
        with testtools.ExpectedException(exceptions.NexusPortBindingNotFound):
            nexus_db_v2.update_nexusport_binding(npb33.port, 200)


class TestCiscoNexusVpcAllocDbTest(testlib_api.SqlTestCase):

    """Unit tests for Cisco mechanism driver's Nexus vpc alloc database."""

    def _make_vpc_list(self, from_in, to_in):

        new_list = []
        for x in range(from_in, (to_in + 1)):
            new_list.append(x)
        return new_list

    def setUp(self):
        original_intersect = nexus_db_v2._get_free_vpcids_on_switches

        def new_get_free_vpcids_on_switches(nexus_ips):
            intersect = list(original_intersect(nexus_ips))
            intersect.sort()
            return intersect

        mock.patch.object(nexus_db_v2,
                         '_get_free_vpcids_on_switches',
                         new=new_get_free_vpcids_on_switches).start()

        super(TestCiscoNexusVpcAllocDbTest, self).setUp()

    def test_vpcalloc_init(self):

        nexus_ips = ['1.1.1.1', '2.2.2.2', '3.3.3.3']

        for this_ip in nexus_ips:
            nexus_db_v2.init_vpc_entries(
                this_ip, self._make_vpc_list(1001, 1025))
            allocs = nexus_db_v2.get_free_switch_vpc_allocs(this_ip)
            self.assertEqual(len(allocs), 25)

        nexus_db_v2.update_vpc_entry(['1.1.1.1'], 1001, False, True)
        nexus_db_v2.update_vpc_entry(['2.2.2.2'], 1002, False, True)
        nexus_db_v2.update_vpc_entry(['3.3.3.3'], 1003, False, True)

        # Verify this update fails since entry already active
        self.assertRaises(
            exceptions.NexusVPCAllocNotFound,
            nexus_db_v2.update_vpc_entry,
            ['3.3.3.3'], 1003, False, True)

        new_vpcid = nexus_db_v2.alloc_vpcid(nexus_ips)
        self.assertEqual(new_vpcid, 1004)

        nexus_db_v2.free_vpcid_for_switch(1002, '2.2.2.2')
        nexus_db_v2.free_vpcid_for_switch_list(1004, nexus_ips)

        # verify vpc 1002 can now be reused
        new_vpcid = nexus_db_v2.alloc_vpcid(nexus_ips)
        self.assertEqual(new_vpcid, 1002)

    def test_vpcalloc_rollback(self):

        nexus_ips = ['1.1.1.1', '2.2.2.2', '3.3.3.3']

        for this_ip in nexus_ips:
            nexus_db_v2.init_vpc_entries(
                this_ip, self._make_vpc_list(1001, 1025))

        nexus_db_v2.update_vpc_entry(
            nexus_ips, 1001, False, True)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('1.1.1.1')
        self.assertEqual(len(allocs), 24)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('2.2.2.2')
        self.assertEqual(len(allocs), 24)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('3.3.3.3')
        self.assertEqual(len(allocs), 24)

        nexus_db_v2.update_vpc_entry(
            nexus_ips, 1001, False, False)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('1.1.1.1')
        self.assertEqual(len(allocs), 25)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('2.2.2.2')
        self.assertEqual(len(allocs), 25)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('3.3.3.3')
        self.assertEqual(len(allocs), 25)

        nexus_db_v2.update_vpc_entry(['3.3.3.3'], 1001, False, True)
        try:
            nexus_db_v2.update_vpc_entry(
                nexus_ips, 1001, False, True)
        except exceptions.NexusVPCAllocNotFound:
            allocs = nexus_db_v2.get_free_switch_vpc_allocs('1.1.1.1')
            self.assertEqual(len(allocs), 25)
            allocs = nexus_db_v2.get_free_switch_vpc_allocs('2.2.2.2')
            self.assertEqual(len(allocs), 25)
            allocs = nexus_db_v2.get_free_switch_vpc_allocs('3.3.3.3')
            self.assertEqual(len(allocs), 24)

    def test_vpcalloc_test_alloc_collision(self):

        def new_get_free_vpcids_on_switches(nexus_ips):
            results = nexus_db_v2.get_free_switch_vpc_allocs('4.4.4.4')
            return results

        nexus_ips = ['1.1.1.1', '2.2.2.2', '3.3.3.3']

        for this_ip in nexus_ips:
            nexus_db_v2.init_vpc_entries(
                this_ip, self._make_vpc_list(1001, 1025))
        # IP 4.4.4.4 is added only to return a list of vpc ids
        # in same format as sql will return.
        nexus_db_v2.init_vpc_entries(
                '4.4.4.4', self._make_vpc_list(1001, 1003))
        mock.patch.object(nexus_db_v2,
                         '_get_free_vpcids_on_switches',
                          new=new_get_free_vpcids_on_switches).start()

        # configure '3.3.3.3', vpcid 1001 so alloc_vpcid will fail
        # on 1001 after updating 1.1.1.1 and 2.2.2.2 and rollback
        # occurs.  Then moves onto successfully allocating 1002.
        nexus_db_v2.update_vpc_entry(['3.3.3.3'], 1001, False, True)
        vpc_id = nexus_db_v2.alloc_vpcid(nexus_ips)
        self.assertEqual(vpc_id, 1002)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('1.1.1.1')
        self.assertEqual(len(allocs), 24)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('2.2.2.2')
        self.assertEqual(len(allocs), 24)
        allocs = nexus_db_v2.get_free_switch_vpc_allocs('3.3.3.3')
        self.assertEqual(len(allocs), 23)

    def test_vpcalloc_min_max(self):

        # Initialize 3 switch vpc entries
        nexus_db_v2.init_vpc_entries(
            '1.1.1.1', self._make_vpc_list(1001, 2000))
        nexus_db_v2.init_vpc_entries(
            '2.2.2.2', self._make_vpc_list(2001, 3000))
        nexus_db_v2.init_vpc_entries(
            '3.3.3.3', self._make_vpc_list(3001, 4000))

        # Verify get_switch_vpc_count_min_max() returns correct
        # count, min, max values for all 3 switches.
        count, min, max = nexus_db_v2.get_switch_vpc_count_min_max(
            '1.1.1.1')
        self.assertEqual(count, 1000)
        self.assertEqual(min, 1001)
        self.assertEqual(max, 2000)

        count, min, max = nexus_db_v2.get_switch_vpc_count_min_max(
            '2.2.2.2')
        self.assertEqual(count, 1000)
        self.assertEqual(min, 2001)
        self.assertEqual(max, 3000)

        count, min, max = nexus_db_v2.get_switch_vpc_count_min_max(
            '3.3.3.3')
        self.assertEqual(count, 1000)
        self.assertEqual(min, 3001)
        self.assertEqual(max, 4000)


class TestCiscoNexusHostMappingDbTest(testlib_api.SqlTestCase):

    """Tests for Nexus Mechanism driver Host Mapping database."""

    def test_enet_host_mapping_db(self):

        nexus_db_v2.add_host_mapping(
            "host-1", "110.1.1.1", "ethernet:1/1", 0, False)
        nexus_db_v2.add_host_mapping(
            "host-1", "112.2.2.2", "ethernet:1/1", 0, False)
        nexus_db_v2.add_host_mapping(
            "host-2", "110.1.1.1", "ethernet:2/2", 0, False)
        nexus_db_v2.add_host_mapping(
            "host-3", "113.3.3.3", "ethernet:3/3", 0, True)
        nexus_db_v2.add_host_mapping(
            "host-4", "114.4.4.4", "ethernet:4/4", 0, True)

        # Do a get 110.1.1.1 and verify  only host-1 is returned
        mappings = nexus_db_v2.get_switch_if_host_mappings(
            "110.1.1.1", "ethernet:1/1")
        self.assertEqual(
            len(mappings),
            1,
            "Unexpected number of switch interface mappings")
        for map in mappings:
            self.assertEqual(
                map.host_id,
                "host-1",
                "Expecting host-1 returned from "
                "get_switch_if_host_mappings")

        # Do a get on host-1 and verify 2 entries returned
        mappings = nexus_db_v2.get_host_mappings("host-1")
        self.assertEqual(
            len(mappings),
            2,
            "Unexpected number of host mappings")
        for map in mappings:
            self.assertEqual(
                map.host_id,
                "host-1",
                "Expecting host-1 returned from "
                "get_host_mappings")
            self.assertEqual(
                map.if_id,
                "ethernet:1/1",
                "Expecting interface returned from "
                "get_host_mappings")

        # Do a get on switch 110.1.1.1 and verify 2 entries returned
        mappings = nexus_db_v2.get_switch_host_mappings("110.1.1.1")
        self.assertEqual(
            len(mappings),
            2,
            "Unexpected number of switch mappings")
        for map in mappings:
            self.assertEqual(
                map.switch_ip,
                "110.1.1.1",
                "Expecting switch_ip returned from "
                "get_switch_host_mappings")

        # Update host mapping by changing the ch_grp
        nexus_db_v2.update_host_mapping(
            "host-2",
            "ethernet:2/2",
            "110.1.1.1",
            2)
        mappings = nexus_db_v2.get_host_mappings("host-2")
        self.assertEqual(
            len(mappings),
            1,
            "Unexpected number of host mappings aft update")
        for map in mappings:
            self.assertEqual(
                map.host_id,
                "host-2",
                "Expecting host-2 returned from "
                "get_host_mappings")
            self.assertEqual(
                map.ch_grp,
                2,
                "Expecting ch_grp 2 returned from "
                "get_host_mappings for host 2")

        # remove 1 host mapping
        nexus_db_v2.remove_host_mapping(
            "ethernet:2/2", "110.1.1.1")
        # Verify it is gone
        self.assertRaises(
            exceptions.NexusHostMappingNotFound,
            nexus_db_v2.get_host_mappings,
            "host-2")

        # remove all static host mapping
        nexus_db_v2.remove_all_static_host_mappings()
        # Verify it is gone
        mappings = nexus_db_v2.get_all_host_mappings()
        self.assertEqual(
            len(mappings),
            2,
            "Unexpected number of non-static entries")
        for map in mappings:
            self.assertFalse(
                map.is_static,
                "Expecting remaining hosts from"
                "get_all_host_mappings to be dynamic")

        # remove host mappings
        nexus_db_v2.remove_host_mapping(
            "ethernet:1/1", "112.2.2.2")
        nexus_db_v2.remove_host_mapping(
            "ethernet:1/1", "110.1.1.1")
        # Verify it is gone
        self.assertRaises(
            exceptions.NexusHostMappingNotFound,
            nexus_db_v2.get_host_mappings,
            "host-1")

    def test_portchannel_host_mapping_db(self):

        nexus_db_v2.add_host_mapping(
            "host-1", "110.1.1.1", "port-channel:100", 0, True)
        nexus_db_v2.add_host_mapping(
            "host-1", "112.2.2.2", "port-channel:100", 0, True)
        nexus_db_v2.add_host_mapping(
            "host-2", "110.1.1.1", "port-channel:100", 0, True)
        nexus_db_v2.add_host_mapping(
            "host-3", "110.1.1.1", "port-channel:100", 0, False)

        # Non-static config should raise DBDuplicateEntry when
        # it already exists.
        self.assertRaises(
            db_exc.DBDuplicateEntry,
            nexus_db_v2.add_host_mapping,
            "host-3", "110.1.1.1", "port-channel:100", 0, False)

        # Static config should NOT raise DBDuplicateEntry when
        # it already exists.
        nexus_db_v2.add_host_mapping(
            "host-2", "110.1.1.1", "port-channel:100", 0, True)

        # Do a get 110.1.1.1 and verify  correct host ids returned
        mappings = nexus_db_v2.get_switch_if_host_mappings(
            "110.1.1.1", "port-channel:100")
        self.assertEqual(
            len(mappings),
            3,
            "Expected 3 switch interface mappings for 110.1.1.1")

        test_host_id = ["host-1", "host-2", "host-3"]
        for map in mappings:
            if map.host_id not in test_host_id:
                raise Exception("Unexpected host from "
                                "get_switch_if_host_mappings")

        # Do a get on host-1 and verify 2 entries returned
        mappings = nexus_db_v2.get_host_mappings("host-1")
        self.assertEqual(
            len(mappings),
            2,
            "Unexpected number of host mappings")
        for map in mappings:
            self.assertEqual(
                map.host_id,
                "host-1",
                "Expecting host-1 returned from "
                "get_host_mappings")
            self.assertEqual(
                map.if_id,
                "port-channel:100",
                "Expecting interface returned from "
                "get_host_mappings")

        # Do a get on switch 110.1.1.1 and verify 2 entries returned
        mappings = nexus_db_v2.get_switch_host_mappings("110.1.1.1")
        self.assertEqual(
            len(mappings),
            3,
            "Expected 3 switch mappings")
        for map in mappings:
            self.assertEqual(
                map.switch_ip,
                "110.1.1.1",
                "Expecting switch_ip returned from "
                "get_switch_host_mappings")

        # Update host mapping by changing the ch_grp
        nexus_db_v2.update_host_mapping(
            "host-2",
            "port-channel:100",
            "110.1.1.1",
            2)
        mappings = nexus_db_v2.get_host_mappings("host-2")
        self.assertEqual(
            len(mappings),
            1,
            "Unexpected number of host mappings aft update")
        for map in mappings:
            self.assertEqual(
                map.host_id,
                "host-2",
                "Expecting host-2 returned from "
                "get_host_mappings")
            self.assertEqual(
                map.ch_grp,
                2,
                "Expecting ch_grp 2 returned from "
                "get_host_mappings for host 2")

        # Not testing remove_host_mapping like above
        # since this is used for baremeatl only in which
        # it only removes ethernet interfaces.

        # remove all static host mapping
        nexus_db_v2.remove_all_static_host_mappings()
        # Verify it is gone
        mappings = nexus_db_v2.get_all_host_mappings()
        self.assertEqual(
            len(mappings),
            1,
            "Unexpected number of non-static entries")
        for map in mappings:
            self.assertFalse(
                map.is_static,
                "Expecting remaining hosts from"
                "get_all_host_mappings to be dynamic")

        # remove host mappings
        nexus_db_v2.remove_host_mapping(
            "port-channel:100", "112.2.2.2")
        # Verify it is gone
        self.assertRaises(
            exceptions.NexusHostMappingNotFound,
            nexus_db_v2.get_host_mappings,
            "host-1")



# Copyright 2017 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""This file contains the public API's for interacting with LLDPAD. """

from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import dfa_sys_lib as utils

LOG = logging.getLogger(__name__)


class LldpApi(object):

    """LLDP API Class. """

    def __init__(self, root_helper):
        self.root_helper = root_helper

    def enable_lldp(self, port_name, is_ncb=True, is_nb=False):
        """Function to enable LLDP on the interface. """
        reply = None
        if is_ncb:
            reply = self.run_lldptool(["-L", "-i", port_name, "-g", "ncb",
                                       "adminStatus=rxtx"])
        elif is_nb:
            reply = self.run_lldptool(["-L", "-i", port_name, "-g", "nb",
                                       "adminStatus=rxtx"])
        else:
            LOG.error("Both NCB and NB are not selected to "
                      "enable LLDP")
            return False
        if reply is None:
            return False
        exp_str = "adminstatus=rxtx"
        if exp_str in reply.replace(" ", "").lower():
            return True
        else:
            return False

    def get_lldp_tlv(self, port_name, is_ncb=True, is_nb=False):
        """Function to Query LLDP TLV on the interface. """
        reply = None
        if is_ncb:
            reply = self.run_lldptool(["get-tlv", "-n", "-i", port_name,
                                       "-g", "ncb"])
        elif is_nb:
            reply = self.run_lldptool(["get-tlv", "-n", "-i", port_name,
                                       "-g", "nb"])
        else:
            LOG.error("Both NCB and NB are not selected to "
                      "query LLDP")
        return reply

    def run_lldptool(self, args):
        """Function for invoking the lldptool utility. """
        full_args = ['lldptool'] + args
        try:
            return utils.execute(full_args, root_helper=self.root_helper)
        except Exception as exc:
            LOG.error("Unable to execute %(cmd)s. "
                      "Exception: %(exception)s",
                      {'cmd': full_args, 'exception': str(exc)})

    def _check_common_tlv_format(self, tlv_complete_data, tlv_data_pattern,
                                 tlv_string):
        """Check for the common TLV format. """
        if tlv_complete_data is None:
            return False, None
        tlv_string_split = tlv_complete_data.split(tlv_string)
        if len(tlv_string_split) < 2:
            return False, None
        next_tlv_list = tlv_string_split[1].split('TLV')[0]
        tlv_val_set = next_tlv_list.split(tlv_data_pattern)
        if len(tlv_val_set) < 2:
            return False, None
        return True, tlv_val_set

    def get_remote_evb_cfgd(self, tlv_data):
        """Returns IF EVB TLV is present in the TLV. """
        return self._check_common_tlv_format(
            tlv_data, "mode:", "EVB Configuration TLV")[0]

    def get_remote_evb_mode(self, tlv_data):
        """Returns the EVB mode in the TLV. """
        ret, parsed_val = self._check_common_tlv_format(
            tlv_data, "mode:", "EVB Configuration TLV")
        if not ret:
            return None
        mode_val = parsed_val[1].split()[0].strip()
        return mode_val

    def get_remote_mgmt_addr(self, tlv_data):
        """Returns Remote Mgmt Addr from the TLV. """
        ret, parsed_val = self._check_common_tlv_format(
            tlv_data, "IPv4:", "Management Address TLV")
        if not ret:
            return None
        addr_fam = 'IPv4:'
        addr = parsed_val[1].split('\n')[0].strip()
        return addr_fam + addr

    def get_remote_sys_desc(self, tlv_data):
        """Returns Remote Sys Desc from the TLV. """
        ret, parsed_val = self._check_common_tlv_format(
            tlv_data, "\n", "System Description TLV")
        if not ret:
            return None
        return parsed_val[1].strip()

    def get_remote_sys_name(self, tlv_data):
        """Returns Remote Sys Name from the TLV. """
        ret, parsed_val = self._check_common_tlv_format(
            tlv_data, "\n", "System Name TLV")
        if not ret:
            return None
        return parsed_val[1].strip()

    def get_remote_port(self, tlv_data):
        """Returns Remote Port from the TLV. """
        ret, parsed_val = self._check_common_tlv_format(
            tlv_data, "\n", "Port Description TLV")
        if not ret:
            return None
        return parsed_val[1].strip()

    def get_remote_chassis_id_mac(self, tlv_data):
        """Returns Remote Chassis ID MAC from the TLV. """
        ret, parsed_val = self._check_common_tlv_format(
            tlv_data, "MAC:", "Chassis ID TLV")
        if not ret:
            return None
        mac = parsed_val[1].split('\n')
        return mac[0].strip()

    def get_remote_port_id_mac(self, tlv_data):
        """Returns Remote Port ID MAC from the TLV. """
        ret, parsed_val = self._check_common_tlv_format(
            tlv_data, "MAC:", "Port ID TLV")
        if not ret:
            return None
        mac = parsed_val[1].split('\n')
        return mac[0].strip()

    def get_remote_port_id_local(self, tlv_data):
        """Returns Remote Port ID Local from the TLV. """
        ret, parsed_val = self._check_common_tlv_format(
            tlv_data, "Local:", "Port ID TLV")
        if not ret:
            return None
        local = parsed_val[1].split('\n')
        return local[0].strip()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_config import cfg

from neutron.tests.unit import testlib_api

from networking_cisco.backwards_compatibility import ml2_api as api

from networking_cisco import backwards_compatibility as bc
from networking_cisco.ml2_drivers.ucsm import (
    mech_cisco_ucsm as md)
from networking_cisco.ml2_drivers.ucsm import config as conf
from networking_cisco.ml2_drivers.ucsm import constants as const
from networking_cisco.ml2_drivers.ucsm import exceptions
from networking_cisco.ml2_drivers.ucsm import ucsm_db
from networking_cisco.ml2_drivers.ucsm import ucsm_network_driver
from networking_cisco.tests.unit.ml2_drivers.ucsm import (
    test_cisco_ucsm_common as mocked)


CONF = cfg.CONF

UCSM_IP_ADDRESS_1 = '1.1.1.1'
UCSM_IP_ADDRESS_2 = '2.2.2.2'

VNIC_NORMAL = 'normal'
VNIC_DIRECT = 'direct'
VNIC_MACVTAP = 'macvtap'
VNIC_TYPES = [VNIC_NORMAL, VNIC_DIRECT, VNIC_MACVTAP]
SRIOV_VNIC_TYPES = [VNIC_DIRECT, VNIC_MACVTAP]

SUPPORTED_PCI_DEVS = ["1137:0071", "8086:10c9"]

NETWORK_ID_1 = 1001
NETWORK_NAME = 'test-network'
VLAN_ID_1 = 100
VLAN_ID_2 = 101
PORT_STATE_ACTIVE = bc.constants.PORT_STATUS_ACTIVE
NETWORK_TYPE = 'vlan'
NETWORK_ID = 'test-network'
PORT_NAME = 'port1'
PORT_NAME2 = 'port2'
PORT_ID = '100001'
PORT_ID2 = '100002'
HOST1 = 'Hostname1'
HOST2 = 'Hostname2'

PCI_INFO_BAD_NIC = '1111:2222'
PCI_INFO_INVALID = '1111'

VLAN_SEGMENT = {api.ID: 'vlan_segment_id',
                api.NETWORK_TYPE: 'vlan',
                api.PHYSICAL_NETWORK: 'test_physnet',
                api.SEGMENTATION_ID: VLAN_ID_1}

VXLAN_SEGMENT = {api.ID: 'vlan_segment_id',
                 api.NETWORK_TYPE: 'vxlan',
                 api.PHYSICAL_NETWORK: 'test_physnet',
                 api.SEGMENTATION_ID: VLAN_ID_1}

VLAN_SEGMENTS_BAD = {api.ID: 'vlan_segment_id',
                     api.NETWORK_TYPE: 'vlan',
                     api.PHYSICAL_NETWORK: 'fake_physnet',
                     api.SEGMENTATION_ID: VLAN_ID_2}

VLAN_SEGMENTS_GOOD = [{api.ID: 'vlan_segment_id',
                       api.NETWORK_TYPE: 'vlan',
                       api.PHYSICAL_NETWORK: 'test_physnet',
                       api.SEGMENTATION_ID: VLAN_ID_2}]

UCSM_HOST_DICT = {HOST1: UCSM_IP_ADDRESS_1,
                  HOST2: UCSM_IP_ADDRESS_2}

PORT_PROFILE_1 = 'OS-PP-100'


class FakeNetworkContext(api.NetworkContext):

    """Network context for testing purposes only."""

    def __init__(self, segments):

        self._network_segments = segments

    @property
    def current(self):
        return {'id': NETWORK_ID_1,
                'name': NETWORK_NAME}

    @property
    def original(self):
        return None

    @property
    def network_segments(self):
        return self._network_segments


class FakePortContext(object):

    """Port context for testing purposes only."""

    def __init__(self, name, port_id, vnic_type, profile,
                 network_context):
        self._port = {
            'status': None,
            'id': port_id,
            'name': name,
            # set for _is_supported_deviceowner() to return True
            'device_owner': bc.constants.DEVICE_OWNER_DHCP,
            bc.portbindings.HOST_ID: HOST1,
            bc.portbindings.VNIC_TYPE: vnic_type,
            bc.portbindings.PROFILE: profile
        }
        self._network = network_context
        self._segment = network_context.network_segments[0]
        self.session = bc.get_writer_session()

    @property
    def current(self):
        return self._port

    @property
    def original(self):
        return None

    @property
    def network(self):
        return self._network

    @property
    def segment(self):
        return self._segment

    @property
    def bottom_bound_segment(self):
        return self._segment

    def set_binding(self, segment_id, vif_type, vif_details,
                    status=None):
        self._bound_segment_id = segment_id
        self._bound_vif_type = vif_type
        self._bound_vif_details = vif_details
        self._port['status'] = status


class FakeServiceProfile(object):
    """Fake Service Profile class for testing only."""
    def __init__(self, service_profile):
        self.sp = service_profile
        self.pn_dn = 'org-root/ls-'
        self.dn = 'org-root'
        self.oper_src_templ_name = (
            self.pn_dn + self.dn)
        self.pn_dn = self.pn_dn + self.sp

    def __iter__(self):
        return self

    def __next__(self):
        return self

    def next(self):
        return self.__next__()


class FakeServer(object):
    def __init__(self, server):
        self.name = server


class FakeUcsmHandle(object):
    """Ucsm connection handle for testing purposes only."""

    def __init__(self, port_profile=None, query_dn=None, invalid_classid=None):
        self._port_profile = port_profile
        self._query_dn = query_dn
        self._invalid_classid = invalid_classid
        self._times_called = 0
        self.sp_list = ['org-root/ls-SP1']
        self.sp_list_temp = []

    def query_dn(self, dn):
        self._times_called += 1

        if self._invalid_classid:
            return FakeServer('nope')
        elif self._query_dn:
            return self._query_dn
        elif dn == 'org-root/ls-SP1':
            return FakeServer(HOST1)
        elif self._times_called == 1:
            return None
        elif self._times_called == 2:
            raise Exception("Port profile still in use by VMs.")
        else:
            return self._port_profile

    def query_classid(self, class_id):
        if self._invalid_classid:
            self.sp_list_temp = [FakeServiceProfile('nope'),
                FakeServiceProfile('nope')]
        else:
            self.sp_list_temp = [FakeServiceProfile('SP1'),
                FakeServiceProfile('SP2')]
        return self.sp_list_temp

    def remove_mo(self, p_profile):
        self._port_profile = None

    def commit(self):
        return

    def logout(self):
        return


class TestCiscoUcsmMechDriver(testlib_api.SqlTestCase,
                              mocked.ConfigMixin):

    """Unit tests for Cisco ML2 UCS Manager MD."""

    def setUp(self):
        """Sets up mock Ucs Sdk."""
        super(TestCiscoUcsmMechDriver, self).setUp()
        self.set_up_mocks()

        def new_ucsm_driver_init(mech_instance):
            mech_instance.ucsmsdk = None
            mech_instance.handles = {}
            mech_instance.supported_sriov_vnic_types = SRIOV_VNIC_TYPES
            mech_instance.supported_pci_devs = SUPPORTED_PCI_DEVS
            mech_instance.ucsm_host_dict = UCSM_HOST_DICT
            conf.load_single_ucsm_config()

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          '__init__',
                          new=new_ucsm_driver_init).start()

        self.mech_driver = md.CiscoUcsmMechanismDriver()
        self.mech_driver.initialize()
        self.vif_type = const.VIF_TYPE_802_QBH
        self.db = ucsm_db.UcsmDbModel()
        self.ucsm_driver = ucsm_network_driver.CiscoUcsmDriver()
        self.ucsm_driver.ucsm_db = ucsm_db.UcsmDbModel()
        conf.load_single_ucsm_config()

    def _create_network_context(self):
        segment = {api.SEGMENTATION_ID: "",
                   api.NETWORK_TYPE: "",
                   }
        segment[api.SEGMENTATION_ID] = VLAN_ID_1
        segment[api.NETWORK_TYPE] = 'vlan'

        network_context = FakeNetworkContext([VLAN_SEGMENT])
        return network_context

    def _create_port_context_vmfex(self):
        """Creates port context with valid VM-FEX vendor info."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_type = VNIC_DIRECT
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}
        network_context = FakeNetworkContext([VLAN_SEGMENT])
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        return port_context

    def _create_port_context_bad(self):
        """Creates port context with badly formed vendor info."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_type = VNIC_DIRECT
        profile = {'pci_vendor_info': PCI_INFO_BAD_NIC}
        network_context = FakeNetworkContext([VLAN_SEGMENT])
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        return port_context

    def _create_port_context_sriov(self):
        """Creates port context with valid SR-IOV vendor info."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_type = VNIC_MACVTAP
        profile = {'pci_vendor_info': const.PCI_INFO_INTEL_82599}
        network_context = FakeNetworkContext([VLAN_SEGMENT])
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        return port_context

    def _create_port_context_normal(self):
        """Creates port context with Normal vnic type."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_type = VNIC_NORMAL
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}
        network_context = FakeNetworkContext([VLAN_SEGMENT])
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        return port_context

    def test_port_supported_deviceowner(self):
        """Verifies detection of supported set of device owners for ports."""
        port_context = self._create_port_context_normal()
        port = port_context._port
        supported_owners = [bc.constants.DEVICE_OWNER_ROUTER_HA_INTF,
                            bc.constants.DEVICE_OWNER_DHCP,
                            'compute:nova']
        for owner in supported_owners:
            port['device_owner'] = owner
            self.assertTrue(self.mech_driver._is_supported_deviceowner(port))

    def test_port_unsupported_deviceowner(self):
        """Verifies detection of unsupported device owners for ports."""
        port_context = self._create_port_context_normal()
        port = port_context._port
        unsupported_owners = [bc.constants.DEVICE_OWNER_ROUTER_INTF,
                              bc.constants.DEVICE_OWNER_ROUTER_GW,
                              bc.constants.DEVICE_OWNER_FLOATINGIP,
                              bc.constants.DEVICE_OWNER_ROUTER_SNAT,
                              bc.constants.DEVICE_OWNER_LOADBALANCER,
                              bc.constants.DEVICE_OWNER_LOADBALANCERV2,
                              'controller:foobar']
        for owner in unsupported_owners:
            port['device_owner'] = owner
            self.assertFalse(self.mech_driver._is_supported_deviceowner(port))

    def test_port_supported_status(self):
        """Verifies detection of supported status values for ports."""
        port_context = self._create_port_context_normal()
        port = port_context._port
        port['status'] = bc.constants.PORT_STATUS_ACTIVE
        self.assertTrue(self.mech_driver._is_status_active(port))

    def test_port_unsupported_status(self):
        """Verifies detection of unsupported status values for ports."""
        port_context = self._create_port_context_normal()
        port = port_context._port
        unsupported_states = [bc.constants.PORT_STATUS_BUILD,
                              bc.constants.PORT_STATUS_DOWN,
                              bc.constants.PORT_STATUS_ERROR,
                              bc.constants.PORT_STATUS_NOTAPPLICABLE]
        for state in unsupported_states:
            port['status'] = state
            self.assertFalse(self.mech_driver._is_status_active(port))

    def test_vmfex_vnic_type_and_vendor_info(self):
        """Verifies VM-FEX port is recognized as a supported vendor."""
        port_context = self._create_port_context_vmfex()
        vnic_type = port_context.current.get(bc.portbindings.VNIC_TYPE,
                                             bc.portbindings.VNIC_NORMAL)
        profile = port_context.current.get(bc.portbindings.PROFILE, {})
        supported = self.ucsm_driver.check_vnic_type_and_vendor_info(
            vnic_type, profile)
        self.assertTrue(supported)

    def test_unsupported_vnic_type_and_vendor_info(self):
        """Verifies unsupported pci vendor is rejected."""
        port_context = self._create_port_context_bad()
        vnic_type = port_context.current.get(bc.portbindings.VNIC_TYPE,
                                             bc.portbindings.VNIC_NORMAL)
        profile = port_context.current.get(bc.portbindings.PROFILE, {})
        supported = self.ucsm_driver.check_vnic_type_and_vendor_info(
            vnic_type, profile)
        self.assertFalse(supported)

    def test_sriov_vnic_type_and_vendor_info(self):
        """Verifies SR-IOV port and MACVTAP vnic_type are supported."""
        port_context = self._create_port_context_sriov()
        vnic_type = port_context.current.get(bc.portbindings.VNIC_TYPE,
                                             bc.portbindings.VNIC_NORMAL)
        profile = port_context.current.get(bc.portbindings.PROFILE, {})
        supported = self.ucsm_driver.check_vnic_type_and_vendor_info(
            vnic_type, profile)
        self.assertTrue(supported)

    def test_normal_vnic_type(self):
        """Verifies NORMAL vnic type is not supported."""
        port_context = self._create_port_context_normal()
        vnic_type = port_context.current.get(bc.portbindings.VNIC_TYPE,
                                             bc.portbindings.VNIC_NORMAL)
        profile = port_context.current.get(bc.portbindings.PROFILE, {})
        supported = self.ucsm_driver.check_vnic_type_and_vendor_info(
            vnic_type, profile)
        self.assertFalse(supported)

    def test_validate_vm_fex_port_cisco(self):
        """Verifies port's pci vendor info makes it VM-FEX capable."""
        port_context = self._create_port_context_vmfex()
        profile = port_context.current.get(bc.portbindings.PROFILE, {})
        valid = self.ucsm_driver.is_vmfex_port(profile)
        self.assertTrue(valid)

    def test_validate_vm_fex_port_bad(self):
        """Verifies unsupported pci vendor is not VM-FEX capable."""
        port_context = self._create_port_context_bad()
        profile = port_context.current.get(bc.portbindings.PROFILE, {})
        valid = self.ucsm_driver.is_vmfex_port(profile)
        self.assertFalse(valid)

    def test_validate_vm_fex_port_sriov(self):
        """Verifies valid SR-IOV port is not VM-FEX capable."""
        port_context = self._create_port_context_sriov()
        profile = port_context.current.get(bc.portbindings.PROFILE, {})
        valid = self.ucsm_driver.is_vmfex_port(profile)
        # For ex: Intel PCI is supported but is not vm-fex.
        # so, should return False
        self.assertFalse(valid)

    def test_check_segment_vlan(self):
        """Verifies VLAN network segments are supported."""
        self.assertTrue(self.mech_driver.check_segment(VLAN_SEGMENT))

    def test_check_segment_vxlan(self):
        """Verifies VXLAN network segments are not supported."""
        self.assertFalse(self.mech_driver.check_segment(VXLAN_SEGMENT))

    def test_vmfex_update_port_precommit(self):
        """Verifies MD saves relevant info for VM-FEX ports into DB."""
        name = PORT_NAME2
        port_id = PORT_ID
        vnic_type = VNIC_DIRECT
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}
        profile_name = "OS-PP-100"

        network_context = self._create_network_context()
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        # Port Profile name and Vlan id are written to DB.
        self.mech_driver.update_port_precommit(port_context)
        # Look for presence of above entry in the DB.
        p_profile = self.db.get_port_profile_for_vlan(VLAN_ID_1,
            UCSM_IP_ADDRESS_1)
        self.assertEqual(profile_name, p_profile)
        # Look to see if flag is set for update_port_postcommit to
        # create Port Profile on UCS Manager.
        self.assertFalse(self.db.is_port_profile_created(VLAN_ID_1,
            UCSM_IP_ADDRESS_1))

    def test_sriov_update_port_precommit(self):
        """Verifies MD does not create Port Profiles for SR-IOV ports."""
        port_context = self._create_port_context_sriov()
        self.mech_driver.update_port_precommit(port_context)
        p_profile = self.db.get_port_profile_for_vlan(VLAN_ID_1,
            UCSM_IP_ADDRESS_1)
        self.assertIsNone(p_profile)

    def test_get_physnet(self):
        expected_physnet = 'test_physnet'
        port_context = self._create_port_context_normal()
        physnet = self.mech_driver._get_physnet(port_context)
        self.assertEqual(expected_physnet, physnet)

    def test_virtio_update_port_precommit(self):
        """Verifies MD adds VNIC Template to DB for Neutron virtio ports."""
        TEST_VNIC_TEMPLATE = 'Test-VNIC'
        TEST_PHYSNET = 'test_physnet'
        port_context = self._create_port_context_normal()

        # Override the vnic template list for this test
        ucsm = CONF.ml2_cisco_ucsm.ucsms[UCSM_IP_ADDRESS_1]
        templates = {TEST_PHYSNET: conf.UCSTemplate('org-root', 'Test-VNIC')}
        CONF.set_override("vnic_template_list", templates, group=ucsm._group)

        vnic_template = ucsm.vnic_template_list.get(TEST_PHYSNET)

        self.assertEqual(TEST_VNIC_TEMPLATE, vnic_template.name)
        self.mech_driver.update_port_precommit(port_context)
        db_entry = self.db.get_vnic_template_vlan_entry(VLAN_ID_1,
                                                        TEST_VNIC_TEMPLATE,
                                                        UCSM_IP_ADDRESS_1,
                                                        TEST_PHYSNET)
        self.assertIsNotNone(db_entry)
        self.assertEqual(VLAN_ID_1, db_entry.vlan_id)

    @mock.patch.object(ucsm_db.UcsmDbModel, 'delete_sp_template_for_vlan')
    @mock.patch.object(ucsm_db.UcsmDbModel, 'delete_vnic_template_for_vlan')
    @mock.patch.object(ucsm_db.UcsmDbModel, 'delete_vlan_entry')
    def test_delete_network_precommit_no_segments(
            self, mock_delete_vlan_entry, mock_delete_vnic,
            mock_delete_sp_template):
        self.mech_driver.delete_network_precommit(FakeNetworkContext([]))
        self.assertFalse(mock_delete_vlan_entry.called)
        self.assertFalse(mock_delete_vnic.called)
        self.assertFalse(mock_delete_sp_template.called)

    @mock.patch.object(ucsm_db.UcsmDbModel, 'delete_sp_template_for_vlan')
    @mock.patch.object(ucsm_db.UcsmDbModel, 'delete_vnic_template_for_vlan')
    @mock.patch.object(ucsm_db.UcsmDbModel, 'delete_vlan_entry')
    def test_delete_network_precommit_vlan_segment(
            self, mock_delete_vlan_entry, mock_delete_vnic,
            mock_delete_sp_template):
        network_context = self._create_network_context()
        vlan_id = network_context.network_segments[0]['segmentation_id']
        self.mech_driver.delete_network_precommit(network_context)
        mock_delete_vlan_entry.assert_called_once_with(vlan_id)
        mock_delete_vnic.assert_called_once_with(vlan_id)
        mock_delete_sp_template.assert_called_once_with(vlan_id)

    def test_update_port_postcommit_success(self):
        """Verifies duplicate Port Profiles are not being created."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_type = VNIC_DIRECT
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}

        network_context = self._create_network_context()
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        self.mech_driver.bind_port(port_context)

        # Port Profile is added to DB and created on UCS Manager.
        self.mech_driver.update_port_precommit(port_context)
        self.assertFalse(self.db.is_port_profile_created(VLAN_ID_1,
            UCSM_IP_ADDRESS_1))

        # Call to UCS Manager driver top level method to create Port Profile
        # is mocked to a new method here. This method verifies input params
        # are correct.
        def new_create_portprofile(mech_context, profile_name, vlan_id,
                                   vnic_type, ucsm_ip, trunk_vlans):
            return True

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'create_portprofile',
                          new=new_create_portprofile).start()

        self.mech_driver.update_port_postcommit(port_context)
        self.assertTrue(self.db.is_port_profile_created(VLAN_ID_1,
            UCSM_IP_ADDRESS_1))

    def test_update_port_postcommit_failure(self):
        """Verifies duplicate Port Profiles are not being created."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_type = VNIC_DIRECT
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}

        network_context = self._create_network_context()
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        self.mech_driver.bind_port(port_context)

        # Port Profile is added to DB and created on UCS Manager.
        self.mech_driver.update_port_precommit(port_context)
        self.assertFalse(self.db.is_port_profile_created(VLAN_ID_1,
            UCSM_IP_ADDRESS_1))

        # Call to UCS Manager driver top level method to create Port Profile
        # is mocked to a new method here. This method verifies input params
        # are correct.
        def new_create_portprofile(mech_context, profile_name, vlan_id,
                                   vnic_type, ucsm_ip, trunk_vlans):
            return False

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'create_portprofile',
                          new=new_create_portprofile).start()

        self.mech_driver.update_port_postcommit(port_context)
        self.assertFalse(self.db.is_port_profile_created(VLAN_ID_1,
             UCSM_IP_ADDRESS_1))

    def test_update_port_postcommit_direct(self):
        """Verifies UCS Manager driver is called with correct parameters."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_direct = VNIC_DIRECT
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}

        network_context = self._create_network_context()
        port_context = FakePortContext(name, port_id, vnic_direct,
                                       profile, network_context)
        self.mech_driver.bind_port(port_context)

        self.mech_driver.update_port_precommit(port_context)

        # Call to UCS Manager driver top level method to create Port Profile
        # is mocked to a new method here. This method verifies input params
        # are correct.
        def new_create_portprofile(mech_context, profile_name, vlan_id,
                                   vnic_type, ucsm_ip, trunk_vlans):
            self.assertEqual("OS-PP-100", profile_name)
            self.assertEqual(100, vlan_id)
            self.assertEqual(VNIC_DIRECT, vnic_type)

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'create_portprofile',
                          new=new_create_portprofile).start()

        self.mech_driver.update_port_postcommit(port_context)

    def test_update_port_postcommit_macvtap(self):
        """Verifies UCS Manager driver is called with correct parameters."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_macvtap = VNIC_MACVTAP
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}

        network_context = self._create_network_context()
        port_context = FakePortContext(name, port_id, vnic_macvtap,
                                       profile, network_context)
        self.mech_driver.bind_port(port_context)

        self.mech_driver.update_port_precommit(port_context)

        # Call to UCS Manager driver top level method to create Port Profile
        # is mocked to a new method here. This method verifies input params
        # are correct.
        def new_create_portprofile(mech_context, profile_name, vlan_id,
                                   vnic_type, ucsm_ip, trunk_vlans):
            self.assertEqual("OS-PP-100", profile_name)
            self.assertEqual(100, vlan_id)
            self.assertEqual(VNIC_MACVTAP, vnic_type)

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'create_portprofile',
                          new=new_create_portprofile).start()

        self.mech_driver.update_port_postcommit(port_context)

    def test_update_port_postcommit_normal(self):
        """Verifies UCS Manager driver is called with correct parameters."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_type = VNIC_NORMAL
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}

        network_context = self._create_network_context()
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        self.mech_driver.bind_port(port_context)

        self.mech_driver.update_port_precommit(port_context)

        # Call to UCS Manager driver top level method to create Port Profile
        # is mocked to a new method here. This method verifies input params
        # are correct.
        def new_update_serviceprofile(mech_context, host_id, vlan_id):
            return True

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'update_serviceprofile',
                          new=new_update_serviceprofile).start()

        self.mech_driver.update_port_postcommit(port_context)

    def test_vnic_template_db_methods(self):
        """Verifies VNIC Template DB methods."""
        TEST_VNIC_TEMPLATE_1 = 'Test-VNIC1'
        TEST_VNIC_TEMPLATE_2 = 'Test-VNIC2'
        TEST_PHYSNET_1 = 'test_physnet1'
        TEST_PHYSNET_2 = 'test_physnet2'
        self.db.add_vnic_template(VLAN_ID_1, UCSM_IP_ADDRESS_1,
                                  TEST_VNIC_TEMPLATE_1, TEST_PHYSNET_1)
        self.db.add_vnic_template(VLAN_ID_2, UCSM_IP_ADDRESS_2,
                                  TEST_VNIC_TEMPLATE_2, TEST_PHYSNET_2)

        db_entry1 = self.db.get_vnic_template_vlan_entry(VLAN_ID_1,
                                                         TEST_VNIC_TEMPLATE_1,
                                                         UCSM_IP_ADDRESS_1,
                                                         TEST_PHYSNET_1)
        self.assertIsNotNone(db_entry1)
        self.assertEqual(VLAN_ID_1, db_entry1.vlan_id)
        self.assertFalse(db_entry1.updated_on_ucs)

        self.db.set_vnic_template_updated(VLAN_ID_2, UCSM_IP_ADDRESS_2,
            TEST_VNIC_TEMPLATE_2, TEST_PHYSNET_2)
        db_entry2 = self.db.get_vnic_template_vlan_entry(VLAN_ID_2,
                                                         TEST_VNIC_TEMPLATE_2,
                                                         UCSM_IP_ADDRESS_2,
                                                         TEST_PHYSNET_2)
        self.assertIsNotNone(db_entry2)
        self.assertEqual(VLAN_ID_2, db_entry2.vlan_id)
        self.assertTrue(db_entry2.updated_on_ucs)

        self.db.delete_vnic_template_for_vlan(VLAN_ID_2)
        db_entry3 = self.db.get_vnic_template_vlan_entry(VLAN_ID_2,
                                                         TEST_VNIC_TEMPLATE_2,
                                                         UCSM_IP_ADDRESS_2,
                                                         TEST_PHYSNET_2)
        self.assertIsNone(db_entry3)

    def test_update_port_postcommit_vnic_template(self):
        """Verifies UCSM driver works correcly with VNIC Templates."""
        TEST_VNIC_TEMPLATE = 'Test-VNIC'
        TEST_PHYSNET = 'test_physnet'
        port_context = self._create_port_context_normal()

        # Override the vnic template list for this test
        ucsm = CONF.ml2_cisco_ucsm.ucsms[UCSM_IP_ADDRESS_1]
        templates = {TEST_PHYSNET: conf.UCSTemplate('org-root', 'Test-VNIC')}
        CONF.set_override("vnic_template_list", templates, group=ucsm._group)

        self.mech_driver.bind_port(port_context)

        physnet = self.mech_driver._get_physnet(port_context)
        self.assertEqual(TEST_PHYSNET, physnet)

        vnic_template = ucsm.vnic_template_list.get(TEST_PHYSNET)

        self.assertIsNotNone(vnic_template)
        self.assertEqual(TEST_VNIC_TEMPLATE, vnic_template.name)

        self.mech_driver.update_port_precommit(port_context)

        def new_update_vnic_template(mech_context, host_id, vlan_id, physnet,
            vnic_template_path, vnic_template):
            return True

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'update_vnic_template',
                          new=new_update_vnic_template).start()

        ucsm_ip = self.ucsm_driver.get_ucsm_ip_for_host(HOST1)
        self.assertEqual(UCSM_IP_ADDRESS_1, ucsm_ip)

        self.mech_driver.update_port_postcommit(port_context)

        db_entry = self.db.get_vnic_template_vlan_entry(VLAN_ID_1,
                                                        vnic_template.name,
                                                        UCSM_IP_ADDRESS_1,
                                                        TEST_PHYSNET)
        self.assertIsNotNone(db_entry)
        self.assertEqual(UCSM_IP_ADDRESS_1, db_entry.device_id)

    def test_bind_port_active(self):
        """Verifies bind_port sets the port status as active."""
        name = PORT_NAME
        port_id = PORT_ID
        vnic_type = VNIC_DIRECT
        profile = {'pci_vendor_info': const.PCI_INFO_CISCO_VIC_1240}

        network_context = FakeNetworkContext(VLAN_SEGMENTS_GOOD)
        port_context = FakePortContext(name, port_id, vnic_type,
                                       profile, network_context)
        self.mech_driver.bind_port(port_context)
        self.assertEqual(PORT_STATE_ACTIVE, port_context._port['status'])

    def test_ucs_manager_disconnect_fail(self):
        """Verifies UCS Manager driver is called with correct parameters."""

        handle = None
        ucsm_ip = UCSM_IP_ADDRESS_2
        self.assertRaises(exceptions.UcsmDisconnectFailed,
                          self.ucsm_driver.ucs_manager_disconnect,
                          handle, ucsm_ip)

    def test_generic_create_profile(self):
        """Test to verify duplicate creation exceptions.

        This is a generic test to mimic the behavior of any UCS Manager
        driver function that creates a profile on the UCS Manager. The
        first time the profile is created, the create succeeds if all
        parameters are correct. If we attempt to create it any number
        of times after that, UCS Manager throws an exception. This test
        code mimics that behavior by using counter to keep track of how
        many times 'update_serviceprofile' is being called.
        counter == 0 -> Simulates invalid input, so raise an exception.
        counter == 1 -> Simulates valid inputs and 1st creation request.
        counter > 1 -> Simulates duplicate creation request and results
        in UCS Manager throwing a duplicate creation request.
        """
        def static_vars(**kwargs):
            def decorate(func):
                for k in kwargs:
                    setattr(func, k, kwargs[k])
                return func
            return decorate

        @static_vars(counter=-1)
        def new_create_ucsm_profile(mech_context, host_id, vlan_id):
            new_create_ucsm_profile.counter += 1
            try:
                if new_create_ucsm_profile.counter == 0:
                    raise Exception("Invalid Operation")
                elif new_create_ucsm_profile.counter > 1:
                    raise Exception(const.DUPLICATE_EXCEPTION)
                else:
                    return True
            except Exception as e:
                if const.DUPLICATE_EXCEPTION in str(e):
                    return True
                else:
                    raise exceptions.UcsmConfigFailed(config=vlan_id,
                                            ucsm_ip=UCSM_IP_ADDRESS_1, exc=e)

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'update_serviceprofile',
                          new=new_create_ucsm_profile).start()

        # Results in new_create_ucsm_profile being called with counter=-1
        self.assertRaises(exceptions.UcsmConfigFailed,
                          self.ucsm_driver.update_serviceprofile,
                          HOST1, VLAN_ID_1)

        # Results in new_create_ucsm_profile being called with counter=0
        self.assertTrue(self.ucsm_driver.update_serviceprofile(
                        HOST1, VLAN_ID_1))

        # Results in new_create_ucsm_profile being called with counter=1
        self.assertTrue(self.ucsm_driver.update_serviceprofile(
                        HOST1, VLAN_ID_1))

    def test_parse_virtio_eth_ports(self):
        """Verifies eth_port_list contains a fully-formed path."""
        eth_port_list = (
            CONF.ml2_cisco_ucsm.ucsms['1.1.1.1'].ucsm_virtio_eth_ports)
        self.assertNotIn('eth4', eth_port_list)
        self.assertIn(const.ETH_PREFIX + 'eth4', eth_port_list)

    def test_host_id_to_hostname(self):
        """Verifies extraction of hostname from host-id from Nova."""
        host_id_with_domain1 = 'compute1.cisco.com'
        expected_hostname1 = 'compute1'

        hostname = self.mech_driver._get_host_id(
            host_id_with_domain1)
        self.assertEqual(expected_hostname1, hostname)

        host_id_with_domain2 = 'compute2.localdomain'
        expected_hostname2 = 'compute2'

        hostname = self.mech_driver._get_host_id(
            host_id_with_domain2)
        self.assertEqual(expected_hostname2, hostname)

        host_id3 = 'compute3'
        hostname = self.mech_driver._get_host_id(host_id3)
        self.assertEqual(host_id3, hostname)

    def test_port_profile_delete_table_add(self):
        """Verifies that add and get of 1 PP to delete table works."""
        self.db.add_port_profile_to_delete_table('OS-PP-100', '10.10.10.10')
        self.assertTrue(self.db.has_port_profile_to_delete('OS-PP-100',
            '10.10.10.10'))

    def test_pp_delete_table_add_multiple(self):
        """Verifies that add and get of multiple PPs to delete table works."""
        self.db.add_port_profile_to_delete_table("OS-PP-100", "10.10.10.10")
        self.db.add_port_profile_to_delete_table("OS-PP-200", "10.10.10.10")
        all_pps = self.db.get_all_port_profiles_to_delete()
        for pp in all_pps:
            self.assertEqual("10.10.10.10", pp.device_id)

    def test_remove_port_profile_from_table(self):
        """Verifies that removing entry from PP delete table works."""
        self.db.add_port_profile_to_delete_table("OS-PP-100", "10.10.10.10")
        self.db.remove_port_profile_to_delete("OS-PP-100", "10.10.10.10")
        self.assertFalse(self.db.has_port_profile_to_delete("OS-PP-100",
            "10.10.10.10"))

    def test_remove_non_existent_port_profile_from_table(self):
        """Verifies that removing previously deleted PP works."""
        self.assertIsNone(self.db.remove_port_profile_to_delete(
            "OS-PP-100", "10.10.10.10"))

    def test_port_profile_delete_on_ucsm(self):
        """Verifies that the PP delete retry logic."""
        handle = FakeUcsmHandle(PORT_PROFILE_1)

        # 1st call to delete_port_profile is designed to not find
        # the PP on the UCSM
        self.ucsm_driver._delete_port_profile(
            handle, PORT_PROFILE_1, UCSM_IP_ADDRESS_1)

        # No entry added to the PP delete table
        self.assertFalse(self.ucsm_driver.ucsm_db.has_port_profile_to_delete(
            PORT_PROFILE_1, UCSM_IP_ADDRESS_1))

        # 2nd call to delete_port_profile is designed to raise exception
        self.ucsm_driver._delete_port_profile(
            handle, PORT_PROFILE_1, UCSM_IP_ADDRESS_1)

        # Failed delete results in entry being created in the PP delete table
        self.assertTrue(self.ucsm_driver.ucsm_db.has_port_profile_to_delete(
            PORT_PROFILE_1, UCSM_IP_ADDRESS_1))

    def test_add_sp_template_config_to_db(self):
        """Verifies the SP template dict has been created properly."""
        host_id = HOST1
        ucsm_ip = UCSM_IP_ADDRESS_1
        sp_template_with_path = "/org-root/test/ls-SP-Test"
        sp_template_info = sp_template_with_path.rsplit('/', 1)

        conf.update_sp_template_config(host_id, ucsm_ip,
                                       sp_template_with_path)

        ucsm = CONF.ml2_cisco_ucsm.ucsms[UCSM_IP_ADDRESS_1]
        self.assertEqual(sp_template_info[1],
                         ucsm.sp_template_list[HOST1].name)
        self.assertEqual(sp_template_info[0],
                         ucsm.sp_template_list[HOST1].path)

    def test_get_ucsm_ip_for_host_success(self):
        """Verfies that ucsm_ip to Service Profile mapping is successful."""
        host_id = HOST1
        ucsm_ip = UCSM_IP_ADDRESS_1
        sp_template_with_path = "/org-root/test/ls-SP-Test"

        conf.update_sp_template_config(host_id, ucsm_ip,
                                       sp_template_with_path)
        self.assertEqual(ucsm_ip, self.ucsm_driver.get_ucsm_ip_for_host(
            host_id))

    def test_get_ucsm_ip_for_host_failure(self):
        """Tests that case where UCSM does not control this host."""
        def new_learn_sp_and_template_for_host(mech_instance, host_id):
            return None

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          '_learn_sp_and_template_for_host',
                          new=new_learn_sp_and_template_for_host).start()

        self.assertIsNone(self.ucsm_driver.get_ucsm_ip_for_host('Hostname3'))

    def test_learn_sp_and_template_for_host_exp(self):
        """Tests case where reading config from UCSM generates exception."""
        host_id = HOST1

        def mocked_connect(self, ucsm_ip):
            handle = mock.Mock()
            return handle

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'ucs_manager_connect',
                          new=mocked_connect).start()

        self.assertRaises(exceptions.UcsmConfigReadFailed,
                          self.ucsm_driver._learn_sp_and_template_for_host,
                          host_id)

    def test_learn_sp_and_template_for_host_error(self):
        """Tests case where learning config from UCSM gives diff host.`"""
        host_id = HOST1

        def mocked_connect(self, ucsm_ip):
            handle = FakeUcsmHandle(PORT_PROFILE_1, FakeServer(HOST2))
            return handle

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'ucs_manager_connect',
                          new=mocked_connect).start()

        self.assertIsNone(
            self.ucsm_driver._learn_sp_and_template_for_host(host_id))

    def test_learn_sp_and_template_for_host_success(self):
        """Tests case where learning config from UCSM gives correct host.`"""
        host_id = HOST1
        expected_ucsm_ip = '2.2.2.2'

        def mocked_connect(self, ucsm_ip):
            if ucsm_ip == expected_ucsm_ip:
                handle = FakeUcsmHandle()
            else:
                handle = FakeUcsmHandle(invalid_classid=True)
            return handle

        mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                          'ucs_manager_connect',
                          new=mocked_connect).start()

        actual_ucsm_ip = self.ucsm_driver._learn_sp_and_template_for_host(
            host_id)

        self.assertEqual(expected_ucsm_ip, actual_ucsm_ip)

        # Resetting the ucsm_host_dict value to what the other tests expect.
        self.ucsm_driver.ucsm_host_dict[HOST1] = '1.1.1.1'



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import eventlet
import netaddr
import pprint as pp

from operator import itemgetter
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_utils import excutils
from oslo_utils import importutils
import six

from neutron.common import exceptions as n_exc
from neutron.common import rpc as n_rpc
from neutron.common import topics

from neutron_lib import exceptions as n_lib_exc

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cfg_agent import cfg_exceptions
from networking_cisco.plugins.cisco.cfg_agent.device_drivers import driver_mgr
from networking_cisco.plugins.cisco.cfg_agent import device_status
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerrole

ncc_errors = importutils.try_import('ncclient.transport.errors')

LOG = logging.getLogger(__name__)


N_ROUTER_PREFIX = 'nrouter-'
ROUTER_ROLE_ATTR = routerrole.ROUTER_ROLE_ATTR

# Minimum number of routers to fetch from server at a time on resync.
# Needed to reduce load on server side and to speed up resync on agent side.
SYNC_ROUTERS_MIN_CHUNK_SIZE = 8


class IPAddressMissingException(n_lib_exc.NeutronException):
    message = _("Router port %(port_id)s has no IP address on subnet "
                "%(subnet_id)s.")


class MultipleIPv4SubnetsException(n_lib_exc.NeutronException):
    message = _("There should not be multiple IPv4 subnets %(subnets)s on "
                "router port %(port_id)s")


class RouterInfo(object):
    """Wrapper class around the (neutron) router dictionary.

    Information about the neutron router is exchanged as a python dictionary
    between plugin and config agent. RouterInfo is a wrapper around that dict,
    with attributes for common parameters. These attributes keep the state
    of the current router configuration, and are used for detecting router
    state changes when an updated router dict is received.

    This is a modified version of the RouterInfo class defined in the
    (reference) l3-agent implementation, for use with cisco config agent.
    """

    def __init__(self, router_id, router):
        self.router_id = router_id
        self.ex_gw_port = None
        self._snat_enabled = None
        self._snat_action = None
        self.internal_ports = []
        self.floating_ips = []
        self._router = None
        self.router = router
        self.routes = []
        self.ha_info = router.get('ha_info')

    @property
    def router(self):
        return self._router

    @property
    def id(self):
        return self.router_id

    @property
    def snat_enabled(self):
        return self._snat_enabled

    @router.setter
    def router(self, value):
        self._router = value
        if not self._router:
            return
        # enable_snat by default if it wasn't specified by plugin
        self._snat_enabled = self._router.get('enable_snat', True)

    def router_name(self):
        return N_ROUTER_PREFIX + self.router_id

    @property
    def ha_enabled(self):
        ha_enabled = self.router.get(ha.ENABLED, False)
        return ha_enabled


class CiscoRoutingPluginApi(object):
    """RoutingServiceHelper(Agent) side of the routing RPC API."""

    def __init__(self, topic, host):
        self.host = host
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def get_routers(self, context, router_ids=None, hd_ids=None):
        """Make a remote process call to retrieve the sync data for routers.

        :param context: session context
        :param router_ids: list of  routers to fetch
        :param hd_ids: hosting device ids, only routers assigned to these
                       hosting devices will be returned.
        """
        cctxt = self.client.prepare(version='1.1')
        return cctxt.call(context, 'cfg_sync_routers', host=self.host,
                          router_ids=router_ids, hosting_device_ids=hd_ids)

    def get_router_ids(self, context, router_ids=None, hd_ids=None):
        """Make a remote process call to retrieve scheduled routers ids."""
        cctxt = self.client.prepare(version='1.3')
        return cctxt.call(context, 'get_cfg_router_ids', host=self.host,
                          router_ids=router_ids, hosting_device_ids=hd_ids)

    def get_hardware_router_type_id(self, context):
        """Get the ID for the ASR1k hardware router type."""
        cctxt = self.client.prepare()
        return cctxt.call(context,
                          'get_hardware_router_type_id',
                          host=self.host)

    def update_floatingip_statuses(self, context, router_id, fip_statuses):
        """Make a remote process call to update operational status for one or
        several floating IPs.

        @param context: contains user information
        @param router_id: id of router associated with the floatingips
        @param fip_statuses: dict with floatingip_id as key and status as value
        """
        cctxt = self.client.prepare(version='1.1')
        return cctxt.call(context, 'update_floatingip_statuses_cfg',
                          router_id=router_id, fip_statuses=fip_statuses)

    def send_update_port_statuses(self, context, port_ids, status):
        """Call the pluging to update the port status which updates the DB.

        :param context: contains user information
        :param port_ids: list of ids of the ports associated with the status
        :param status: value of the status for the given port list (port_ids)
        """
        cctxt = self.client.prepare(version='1.1')
        return cctxt.call(context, 'update_port_statuses_cfg',
                          port_ids=port_ids, status=status)


class RoutingServiceHelper(object):

    target = oslo_messaging.Target(version='1.1')

    def __init__(self, host, conf, cfg_agent):
        self.conf = conf
        self.cfg_agent = cfg_agent
        self.context = bc.context.get_admin_context_without_session()
        self.plugin_rpc = CiscoRoutingPluginApi(topics.L3PLUGIN, host)
        self._dev_status = device_status.DeviceStatus()
        self._dev_status.enable_heartbeat = (
            self.conf.cfg_agent.enable_heartbeat)
        self._drivermgr = driver_mgr.DeviceDriverManager(cfg_agent)

        self.router_info = {}
        self.updated_routers = set()
        self.removed_routers = set()
        self.sync_devices = set()
        self.sync_devices_attempts = 0
        self.fullsync = True
        self.sync_routers_chunk_size = (
            cfg.CONF.cfg_agent.max_device_sync_batch_size)
        self.topic = '%s.%s' % (c_constants.CFG_AGENT_L3_ROUTING, host)

        self.hardware_router_type = None
        self.hardware_router_type_id = None

        self._setup_rpc()

    def _setup_rpc(self):
        self.conn = n_rpc.create_connection()
        self.endpoints = [self]
        self.conn.create_consumer(self.topic, self.endpoints, fanout=False)
        self.conn.consume_in_threads()

    ### Notifications from Plugin ####

    def router_deleted(self, context, routers):
        """Deal with router deletion RPC message."""
        LOG.debug('Got router deleted notification for %s', routers)
        self._update_removed_routers_cache(routers)

    def routers_updated(self, context, routers):
        """Deal with routers modification and creation RPC message."""
        LOG.debug('Got routers updated notification :%s', routers)
        if routers:
            # This is needed for backward compatibility
            if isinstance(routers[0], dict):
                routers = [router['id'] for router in routers]
            self._update_updated_routers_cache(routers)

    def router_removed_from_hosting_device(self, context, routers):
        LOG.debug('Got router removed from hosting device: %s', routers)
        self.router_deleted(context, routers)

    def router_added_to_hosting_device(self, context, routers):
        LOG.debug('Got router added to hosting device :%s', routers)
        self.routers_updated(context, routers)

    # version 1.1
    def routers_removed_from_hosting_device(self, context, router_ids):
        LOG.debug('Got routers removed from hosting device: %s', router_ids)
        self.router_deleted(context, router_ids)

    # Routing service helper public methods

    @property
    def driver_manager(self):
        return self._drivermgr

    def process_service(self, device_ids=None, removed_devices_info=None):
        try:
            LOG.debug("Routing service processing started")
            resources = {}
            routers = []
            removed_routers = []
            all_routers_flag = False
            if self.fullsync:
                LOG.debug("The fullsync flag is set. Starting complete sync")
                # Setting all_routers_flag and clear the global full_sync flag
                all_routers_flag = True
                self.fullsync = False
                self.router_info = {}
                self._get_and_clear_updated_routers_cache()
                self._get_and_clear_removed_routers_cache()
                self.sync_devices.clear()
                routers = self._fetch_router_info(all_routers=True)
                LOG.debug("All routers: %s" % (pp.pformat(routers)))
                if routers is not None:
                    self._cleanup_invalid_cfg(routers)
            else:
                if self.updated_routers:
                    router_ids = self._get_and_clear_updated_routers_cache()
                    LOG.debug("Updated routers: %s", router_ids)
                    routers = self._fetch_router_info(router_ids=router_ids)
                    LOG.debug("Updated routers: %s" % (pp.pformat(routers)))
                if device_ids:
                    LOG.debug("Adding new devices: %s", device_ids)
                    self.sync_devices = set(device_ids) | self.sync_devices
                if self.sync_devices:
                    self._handle_sync_devices(routers)
                if removed_devices_info:
                    if removed_devices_info.get('deconfigure'):
                        ids = self._get_router_ids_from_removed_devices_info(
                            removed_devices_info)
                        self._update_removed_routers_cache(ids)
                if self.removed_routers:
                    removed_routers_ids = self._get_removed_routers_cache()
                    LOG.debug("Removed routers: %s",
                              pp.pformat(removed_routers_ids))
                    for r in removed_routers_ids:
                        if r in self.router_info:
                            removed_routers.append(self.router_info[r].router)

            # Sort on hosting device
            if routers:
                resources['routers'] = routers
            if removed_routers:
                resources['removed_routers'] = removed_routers
            hosting_devices = self._sort_resources_per_hosting_device(
                resources)

            # Dispatch process_services() for each hosting device
            pool = eventlet.GreenPool()
            for device_id, resources in hosting_devices.items():
                routers = resources.get('routers', [])
                removed_routers = resources.get('removed_routers', [])
                pool.spawn_n(self._process_routers, routers, removed_routers,
                             device_id, all_routers=all_routers_flag)
            pool.waitall()
            if removed_devices_info:
                for hd_id in removed_devices_info['hosting_data']:
                    self.driver_manager.remove_driver_for_hosting_device(hd_id)
            LOG.debug("Routing service processing successfully completed")
        except Exception:
            LOG.exception("Failed processing routers")
            self.fullsync = True

    def collect_state(self, configurations):
        """Collect state from this helper.

        A set of attributes which summarizes the state of the routers and
        configurations managed by this config agent.
        :param configurations: dict of configuration values
        :return dict of updated configuration values
        """
        num_ex_gw_ports = 0
        num_interfaces = 0
        num_floating_ips = 0
        router_infos = self.router_info.values()
        num_routers = len(router_infos)
        num_hd_routers = collections.defaultdict(int)
        for ri in router_infos:
            ex_gw_port = ri.router.get('gw_port')
            if ex_gw_port:
                num_ex_gw_ports += 1
            num_interfaces += len(ri.router.get(
                bc.constants.INTERFACE_KEY, []))
            num_floating_ips += len(ri.router.get(
                bc.constants.FLOATINGIP_KEY, []))
            hd = ri.router['hosting_device']
            if hd:
                num_hd_routers[hd['id']] += 1
        routers_per_hd = dict((hd_id, {'routers': num})
                              for hd_id, num in num_hd_routers.items())
        non_responding = self._dev_status.get_backlogged_hosting_devices()
        configurations['total routers'] = num_routers
        configurations['total ex_gw_ports'] = num_ex_gw_ports
        configurations['total interfaces'] = num_interfaces
        configurations['total floating_ips'] = num_floating_ips
        configurations['hosting_devices'] = routers_per_hd
        configurations['non_responding_hosting_devices'] = non_responding
        return configurations

    # Routing service helper internal methods

    @lockutils.synchronized('updated_routers_cache', 'cisco-cfg-agent-')
    def _get_and_clear_updated_routers_cache(self):
        router_ids = list(self.updated_routers)
        self.updated_routers.clear()
        return router_ids

    @lockutils.synchronized('updated_routers_cache', 'cisco-cfg-agent-')
    def _update_updated_routers_cache(self, routers):
        self.updated_routers.update(routers)

    @lockutils.synchronized('updated_routers_cache', 'cisco-cfg-agent-')
    def _del_from_updated_routers_cache(self, router_id):
        self.updated_routers.discard(router_id)

    @lockutils.synchronized('removed_routers_cache', 'cisco-cfg-agent-')
    def _get_and_clear_removed_routers_cache(self):
        router_ids = list(self.removed_routers)
        self.removed_routers.clear()
        return router_ids

    @lockutils.synchronized('removed_routers_cache', 'cisco-cfg-agent-')
    def _get_removed_routers_cache(self):
        return list(self.removed_routers)

    @lockutils.synchronized('removed_routers_cache', 'cisco-cfg-agent-')
    def _update_removed_routers_cache(self, routers):
        self.removed_routers.update(routers)

    @lockutils.synchronized('removed_routers_cache', 'cisco-cfg-agent-')
    def _del_from_removed_routers_cache(self, router_id):
        self.removed_routers.discard(router_id)

    def _cleanup_invalid_cfg(self, routers):

        # dict with hd id as key and associated routers list as val
        hd_routermapping = collections.defaultdict(list)
        for router in routers:
            hd_routermapping[router['hosting_device']['id']].append(router)

        # call cfg cleanup specific to device type from its driver
        for hd_id, routers in six.iteritems(hd_routermapping):
            temp_res = {"id": hd_id,
                        "hosting_device": routers[0]['hosting_device'],
                        "router_type": routers[0]['router_type']}
            driver = self.driver_manager.set_driver(temp_res)
            LOG.debug("Running config sync for hosting device %(hd_id)s that "
                      "should host %(num_r)d routers",
                      {'hd_id': hd_id, 'num_r': len(routers)})
            driver.cleanup_invalid_cfg(
                routers[0]['hosting_device'], routers)

    def _fetch_router_info(self, router_ids=None, device_ids=None,
                           all_routers=False):
        """Fetch router dict from the routing plugin.

        :param router_ids: List of router_ids of routers to fetch
        :param device_ids: List of device_ids whose routers to fetch
        :param all_routers:  If True fetch all the routers for this agent.
        :return: List of router dicts of format:
                 [ {router_dict1}, {router_dict2},.....]
        """
        try:
            if all_routers:
                LOG.debug('Fetching all routers')
                router_ids = self.plugin_rpc.get_router_ids(self.context)
                routers = self._fetch_router_chunk_data(router_ids)

            elif router_ids:
                routers = self._fetch_router_chunk_data(router_ids)

            elif device_ids:
                return self.plugin_rpc.get_routers(self.context,
                                                   hd_ids=device_ids)
        except oslo_messaging.MessagingTimeout:
            if self.sync_routers_chunk_size > SYNC_ROUTERS_MIN_CHUNK_SIZE:
                self.sync_routers_chunk_size = max(
                    int(round(self.sync_routers_chunk_size / 2)),
                    SYNC_ROUTERS_MIN_CHUNK_SIZE)
                LOG.warning('Server failed to return info for routers in '
                            'required time, decreasing chunk size to: %s',
                            self.sync_routers_chunk_size)
            else:
                LOG.warning('Server failed to return info for routers in '
                            'required time even with min chunk size: %s. '
                            'It might be under very high load or just '
                            'inoperable',
                            self.sync_routers_chunk_size)
            raise
        except oslo_messaging.MessagingException:
            LOG.exception("RPC Error in fetching routers from plugin")
            self.fullsync = True
            raise n_exc.AbortSyncRouters()

        LOG.debug("Periodic_sync_routers_task successfully completed")
        # adjust chunk size after successful sync
        if (self.sync_routers_chunk_size <
                cfg.CONF.cfg_agent.max_device_sync_batch_size):
            self.sync_routers_chunk_size = min(
                self.sync_routers_chunk_size + SYNC_ROUTERS_MIN_CHUNK_SIZE,
                cfg.CONF.cfg_agent.max_device_sync_batch_size)
        return routers

    def _fetch_router_chunk_data(self, router_ids=None):

        """Fetch router data from the routing plugin in chunks.

                :param router_ids: List of router_ids of routers to fetch
                :return: List of router dicts of format:
                         [ {router_dict1}, {router_dict2},.....]
        """

        curr_router = []
        if len(router_ids) > self.sync_routers_chunk_size:
            # fetch routers by chunks to reduce the load on server and
            # to start router processing earlier
            for i in range(0, len(router_ids),
                           self.sync_routers_chunk_size):
                routers = self.plugin_rpc.get_routers(
                                self.context, (router_ids[i:i +
                                               self.sync_routers_chunk_size]))
                LOG.debug('Processing :%r', routers)
                for r in routers:
                    curr_router.append(r)
        else:
            curr_router = self.plugin_rpc.get_routers(
                self.context, router_ids=router_ids)
        return curr_router

    def _handle_sync_devices(self, routers):
        """
        Handles routers during a device_sync.

        This method performs post-processing on routers fetched from the
        routing plugin during a device sync.  Routers are first fetched
        from the plugin based on the list of device_ids.  Since fetched
        routers take precedence over pending work, matching router-ids
        buffered in update_routers and removed_routers are discarded.
        The existing router cache is also cleared in order to properly
        trigger updates and deletes.  Lastly, invalid configuration in
        the underlying hosting-device is deleted via _cleanup_invalid_cfg.

        Modifies updated_routers, removed_routers, and sync_devices
        attributes

        :param routers: working list of routers as populated in
                        process_services
        """
        sync_devices_list = list(self.sync_devices)
        LOG.debug("Fetching routers on:%s", sync_devices_list)
        fetched_routers = self._fetch_router_info(device_ids=sync_devices_list)

        if fetched_routers:
            LOG.debug("[sync_devices] Fetched routers :%s",
                      pp.pformat(fetched_routers))

            # clear router_config cache
            for router_dict in fetched_routers:
                self._del_from_updated_routers_cache(router_dict['id'])
                self._del_from_removed_routers_cache(router_dict['id'])
                LOG.debug("[sync_devices] invoking "
                          "_router_removed(%s)",
                          router_dict['id'])
                self._router_removed(router_dict['id'],
                                     deconfigure=False)

            self._cleanup_invalid_cfg(fetched_routers)
            routers.extend(fetched_routers)
            self.sync_devices.clear()
            LOG.debug("[sync_devices] %s finished",
                      sync_devices_list)
        else:
            # If the initial attempt to sync a device
            # failed, retry again (by not clearing sync_devices)
            # Normal updated_routers processing is still allowed
            # to happen
            self.sync_devices_attempts += 1

            if (self.sync_devices_attempts >=
                cfg.CONF.cfg_agent.max_device_sync_attempts):

                LOG.debug("Max number [%d / %d ] of sync_devices "
                          "attempted.  No further retries will "
                          "be attempted." %
                          (self.sync_devices_attempts,
                           cfg.CONF.cfg_agent.max_device_sync_attempts))
                self.sync_devices.clear()
                self.sync_devices_attempts = 0
            else:
                LOG.debug("Fetched routers was blank for sync attempt "
                          "[%d / %d], will attempt resync of %s devices "
                          "again in the next iteration" %
                          (self.sync_devices_attempts,
                           cfg.CONF.cfg_agent.max_device_sync_attempts,
                           pp.pformat(self.sync_devices)))

    @staticmethod
    def _get_router_ids_from_removed_devices_info(removed_devices_info):
        """Extract router_ids from the removed devices info dict.

        :param removed_devices_info: Dict of removed devices and their
               associated resources.
        Format:
                {
                  'hosting_data': {'hd_id1': {'routers': [id1, id2, ...]},
                                   'hd_id2': {'routers': [id3, id4, ...]},
                                   ...
                                  },
                  'deconfigure': True/False
                }
        :return removed_router_ids: List of removed router ids
        """
        removed_router_ids = []
        for hd_id, resources in removed_devices_info['hosting_data'].items():
            removed_router_ids += resources.get('routers', [])
        return removed_router_ids

    @staticmethod
    def _sort_resources_per_hosting_device(resources):
        """This function will sort the resources on hosting device.

        The sorting on hosting device is done by looking up the
        `hosting_device` attribute of the resource, and its `id`.

        :param resources: a dict with key of resource name
        :return dict sorted on the hosting device of input resource. Format:
        hosting_devices = {
                            'hd_id1' : {'routers':[routers],
                                        'removed_routers':[routers], .... }
                            'hd_id2' : {'routers':[routers], .. }
                            .......
                            }
        """
        hosting_devices = {}
        for key in resources.keys():
            for r in resources.get(key) or []:
                if r.get('hosting_device') is None:
                    continue
                hd_id = r['hosting_device']['id']
                hosting_devices.setdefault(hd_id, {})
                hosting_devices[hd_id].setdefault(key, []).append(r)
        return hosting_devices

    def _adjust_router_list_for_global_router(self, routers):
        """
        Pushes 'Global' routers to the end of the router list, so that
        deleting default route occurs before deletion of external nw subintf
        """
        #ToDo(Hareesh): Simplify if possible
        for r in routers:
            if r[ROUTER_ROLE_ATTR] == c_constants.ROUTER_ROLE_GLOBAL:
                LOG.debug("Global router:%s found. Moved to the end of list "
                          "for processing", r['id'])
                routers.remove(r)
                routers.append(r)

    def _process_routers(self, routers, removed_routers,
                         device_id=None, all_routers=False):
        """Process the set of routers.

        Iterating on the set of routers received and comparing it with the
        set of routers already in the routing service helper, new routers
        which are added are identified. Before processing check the
        reachability (via ping) of hosting device where the router is hosted.
        If device is not reachable it is backlogged.

        For routers which are only updated, call `_process_router()` on them.

        When all_routers is set to True (because of a full sync),
        this will result in the detection and deletion of routers which
        have been removed.

        Whether the router can only be assigned to a particular hosting device
        is decided and enforced by the plugin. No checks are done here.

        :param routers: The set of routers to be processed
        :param removed_routers: the set of routers which where removed
        :param device_id: Id of the hosting device
        :param all_routers: Flag for specifying a partial list of routers
        :return: None
        """
        try:
            ids_previously_hosted_routers = (
                set(r_id for r_id, rdata in self.router_info.items()
                    if rdata.router.get('hosting_device',
                                        {}).get('id') == device_id))

            if all_routers:
                prev_router_ids = ids_previously_hosted_routers
            else:
                prev_router_ids = (ids_previously_hosted_routers &
                                   set([router['id'] for router in routers]))
            cur_router_ids = set()
            deleted_routerids_list = []

            for r in routers:
                if not r['admin_state_up']:
                        continue
                cur_router_ids.add(r['id'])

            # identify list of routers(ids) that no longer exist
            for router_id in prev_router_ids - cur_router_ids:
                deleted_routerids_list.append(router_id)
            if removed_routers:
                self._adjust_router_list_for_global_router(removed_routers)
                for router in removed_routers:
                    deleted_routerids_list.append(router['id'])

            self._adjust_router_list_for_global_router(routers)
            # First process create/updated routers
            for r in routers:
                LOG.debug("Processing router[id:%(id)s, role:%(role)s]",
                          {'id': r['id'], 'role': r[ROUTER_ROLE_ATTR]})
                if r['id'] in deleted_routerids_list:
                    continue
                if r['status'] == c_constants.ROUTER_INFO_INCOMPLETE:
                    # The plugin could not fill in all the info due to
                    # timing and db settling down. So put this router
                    # back in updated_routers, we will pull again on the
                    # sync time.
                    LOG.debug("Router: %(id)s INFO_INCOMPLETE",
                              {'id': r['id']})
                    self._update_updated_routers_cache([r['id']])
                    continue
                try:
                    if not r['admin_state_up']:
                        continue
                    cur_router_ids.add(r['id'])
                    hd = r['hosting_device']
                    if not self._dev_status.is_hosting_device_reachable(hd):
                        LOG.info("Router: %(id)s is on an unreachable "
                                 "hosting device. ", {'id': r['id']})
                        continue
                    if r['id'] not in self.router_info:
                        self._router_added(r['id'], r)
                    ri = self.router_info[r['id']]
                    ri.router = r
                    self._process_router(ri)
                except ncc_errors.SessionCloseError as e:
                    LOG.exception(
                        "ncclient Unexpected session close %s", e)
                    if not self._dev_status.is_hosting_device_reachable(
                        r['hosting_device']):
                        LOG.debug("Lost connectivity to hosting device %s" %
                                  r['hosting_device']['id'])
                        # Will rely on heartbeat to detect hd state
                        # and schedule resync when hd comes back
                    else:
                        # retry the router update on the next pass
                        self._update_updated_routers_cache([r['id']])
                        LOG.debug("RETRY_RTR_UPDATE %s" % (r['id']))

                    continue
                except KeyError as e:
                    LOG.exception("Key Error, missing key: %s", e)
                    self._update_updated_routers_cache([r['id']])
                    continue
                except cfg_exceptions.DriverException as e:
                    LOG.exception("Driver Exception on router:%(id)s. "
                                  "Error is %(e)s", {'id': r['id'],
                                                     'e': e})
                    self._update_updated_routers_cache([r['id']])
                    continue
                LOG.debug("Done processing router[id:%(id)s, role:%(role)s]",
                          {'id': r['id'], 'role': r[ROUTER_ROLE_ATTR]})
            # Finally process removed routers
            for router_id in deleted_routerids_list:
                LOG.debug("Processing deleted router:%s", router_id)
                self._router_removed(router_id)
        except Exception:
            LOG.exception("Exception in processing routers on device:%s",
                          device_id)
            self.sync_devices.add(device_id)

    def _send_update_port_statuses(self, port_ids, status):
        """Sends update notifications to set the operational status of the
        list of router ports provided. To make each notification doesn't exceed
        the RPC length, each message contains a maximum of MAX_PORTS_IN_BATCH
        port ids.

        :param port_ids: List of ports to update the status
        :param status: operational status to update
                       (ex: bc.constants.PORT_STATUS_ACTIVE)
        """
        if not port_ids:
            return

        MAX_PORTS_IN_BATCH = 50
        list_chunks_ports = [port_ids[i:i + MAX_PORTS_IN_BATCH]
            for i in six.moves.range(0, len(port_ids), MAX_PORTS_IN_BATCH)]
        for chunk_ports in list_chunks_ports:
            self.plugin_rpc.send_update_port_statuses(self.context,
                                                      chunk_ports, status)

    def _get_ports_on_networks(self, router_role, internal_ports):
        # Tenant router and any of its HA redundancy routers may have more
        # than one interface on an internal network (on different subnets).
        # This is not the case for Global routers that are not connected to
        # any internal networks.
        relevant_router_roles = [None, c_constants.ROUTER_ROLE_HA_REDUNDANCY]
        nw_p_dict = {}
        if router_role in relevant_router_roles:
            for p in internal_ports:
                new_item = {'ip_address': p['fixed_ips'][0]['ip_address'],
                            'port': p}
                try:
                    nw_p_dict[p['network_id']].append(new_item)
                except KeyError:
                    nw_p_dict[p['network_id']] = [new_item]
        # To configure primary and secondary ip addresses in a consistent
        # way across the tenant router and any HA redundancy routers we
        # sort ports in each network on their ip address
        multiple_port_networks = {
            k: [vv['port'] for vv in sorted(v, key=itemgetter('ip_address'))]
            for k, v in six.iteritems(nw_p_dict)}
        return multiple_port_networks

    def _get_internal_port_changes(self, ri, internal_ports):
        existing_port_ids = set([p['id'] for p in ri.internal_ports])
        current_port_ids = set([p['id'] for p in internal_ports
                                if p['admin_state_up']])
        new_ports = []
        if existing_port_ids == current_port_ids:
            # Check if a new subnet is connected to the port
            for p in internal_ports:
                new_ports = [p for q in ri.internal_ports if (
                              len(p['fixed_ips']) > len(q['fixed_ips']))]
        else:
            new_ports = [p for p in internal_ports
                     if
                     p['id'] in (current_port_ids - existing_port_ids)]
        old_ports = [p for p in ri.internal_ports
                     if p['id'] not in current_port_ids]

        router_role = ri.router[routerrole.ROUTER_ROLE_ATTR]
        new_ports_on_networks = self._get_ports_on_networks(router_role,
                                                            new_ports)
        current_ports_on_networks = self._get_ports_on_networks(
            router_role, internal_ports)
        old_ports_on_networks = self._get_ports_on_networks(router_role,
                                                            old_ports)
        former_ports_on_networks = (
            self._get_ports_on_networks(router_role, ri.internal_ports))
        change_details = {
            nw_id: {'new_ports': new_ports_on_networks.get(nw_id, []),
                    'current_ports': current_ports_on_networks.get(nw_id, []),
                    'old_ports': old_ports_on_networks.get(nw_id, []),
                    'former_ports': former_ports_on_networks.get(nw_id, [])}
            for nw_id in set(current_ports_on_networks.keys()) |
            set(former_ports_on_networks.keys())}

        LOG.debug("Ids of ports to be added = %s" % pp.pformat(
            [p['id'] for p in new_ports]))
        LOG.debug("Ids of ports to removed = %s" % pp.pformat(
            [p['id'] for p in old_ports]))

        return new_ports, old_ports, change_details

    def _enable_disable_ports(self, ri, ex_gw_port, internal_ports):
        if not ri.router['admin_state_up']:
            self._disable_router_interface(ri)
        else:
            if ex_gw_port:
                if not ex_gw_port['admin_state_up']:
                    self._disable_router_interface(ri, ex_gw_port)
                else:
                    self._enable_router_interface(ri, ex_gw_port)
            for port in internal_ports:
                if not port['admin_state_up']:
                    self._disable_router_interface(ri, port)
                else:
                    self._enable_router_interface(ri, port)

    def _process_new_ports(self, ri, new_ports, ex_gw_port, list_port_ids_up,
                           change_details):
        for p in new_ports:
            num_subnets_on_port = len(p['subnets'])
            if num_subnets_on_port > 1:
                LOG.error("Ignoring router port with multiple IPv4 "
                          "subnets associated")
                raise MultipleIPv4SubnetsException(
                    port_id=p['id'], subnets=pp.pformat(p['subnets']))
            p['change_details'] = change_details[p['network_id']]
            current_ports_on_network = p['change_details']['current_ports']
            if len(current_ports_on_network) > 1:
                # port p is one of at least two ports on the network so we must
                # configure one as primary and the others as secondary
                is_primary = current_ports_on_network[0]['id'] == p['id']
                LOG.debug('Processing new port %(p_id)s with IPv4 %(ip)s '
                          'which should be %(type)s address. The router has '
                          'additional ports on the network (on different '
                          'subnets).',
                          {'p_id': p['id'],
                           'type': 'primary' if is_primary else 'secondary',
                           'ip': p['fixed_ips'][0]['ip_address']})
            else:
                # only one port on network so is must be the primary
                is_primary = True
                LOG.debug('Processing new port %(p_id)s with IPv4 %(ip)s '
                          'which should be primary address as router has no '
                          'other ports on network',
                          {'p_id': p['id'],
                           'ip': p['fixed_ips'][0]['ip_address']})
            self._set_subnet_info(p, p['subnets'][0]['id'], is_primary)
            self._internal_network_added(ri, p, ex_gw_port)

            if not any(i_p['id'] == p['id'] for i_p in ri.internal_ports):
                ri.internal_ports.append(p)
            list_port_ids_up.append(p['id'])

    def _process_new_ports_global(self, ri, new_ports, ex_gw_port,
                                  list_port_ids_up):
        #TODO(bmelande): 1. We need to handle the case where an external
        #                   network, to which a global router is connected,
        #                   is given another subnet. The global router must
        #                   then attached to that subnet. That attachment
        #                   does NOT result in a new router port. Instead, it
        #                   is done as an update to an EXISTING router port
        #                   which gets another ip address (from the newly
        #                   added subnet.
        for p in new_ports:
            # We sort the port's subnets on subnet_id so we can be sure that
            # the same ip address is used as primary on the HA master router
            # as well as on all HA backup routers.
            port_subnets = sorted(p['subnets'], key=itemgetter('id'))
            num_subnets_on_port = len(port_subnets)
            LOG.debug("Number of subnets associated with new Global router "
                      "port = %d", num_subnets_on_port)
            # Configure the primary IP address
            self._set_subnet_info(p, port_subnets[0]['id'])
            self._internal_network_added(ri, p, ex_gw_port)

            # Process the secondary subnets. Only router ports of global
            # routers can have multiple ipv4 subnets since we connect such
            # routers to external networks using regular router ports.
            for p_sn in port_subnets[1:]:
                self._set_subnet_info(p, p_sn['id'], is_primary=False)
                self._internal_network_added(ri, p, ex_gw_port)

            ri.internal_ports.append(p)
            list_port_ids_up.append(p['id'])

    def _process_old_ports(self, ri, old_ports, ex_gw_port, change_details):
        for p in old_ports:
            p['change_details'] = change_details[p['network_id']]
            former_ports_on_network = p['change_details']['former_ports']
            if len(former_ports_on_network) > 1:
                # port p is one of at least two ports on the network
                # so we must deconfigure it accordingly
                is_primary = former_ports_on_network[0]['id'] == p['id']
                LOG.debug('Processing deletion of port %(p_id)s with IPv4 '
                          '%(ip)s which was %(type)s address. The router has '
                          'additional ports on the network (on different '
                          'subnets).',
                          {'p_id': p['id'],
                           'type': 'primary' if is_primary else 'secondary',
                           'ip': p['fixed_ips'][0]['ip_address']})
            else:
                # only one port on network so is must be the primary
                is_primary = True
                LOG.debug('Processing deletion of port %(p_id)s with IPv4 '
                          '%(ip)s which was primary address as router had no '
                          'other ports on network',
                          {'p_id': p['id'],
                           'ip': p['fixed_ips'][0]['ip_address']})
            self._set_subnet_info(p, p['subnets'][0]['id'], is_primary)
            self._internal_network_removed(ri, p, ri.ex_gw_port)
            ri.internal_ports.remove(p)

    def _process_old_ports_global(self, ri, old_ports, ex_gw_port):
        for p in old_ports:
            self._internal_network_removed(ri, p, ri.ex_gw_port)
            ri.internal_ports.remove(p)

    def _process_gateway_set(self, ri, ex_gw_port, list_port_ids_up):
        # We sort the port's subnets on subnet_id so we can be sure that
        # the same ip address is used as primary on the HA master router
        # as well as on all HA backup routers.
        gw_port_subnets = sorted(ex_gw_port['subnets'], key=itemgetter('id'))

        # Configure the primary IP address
        self._set_subnet_info(ex_gw_port, gw_port_subnets[0]['id'])
        self._external_gateway_added(ri, ex_gw_port)

        # Process the secondary subnets
        for gw_p_sn in gw_port_subnets[1:]:
            self._set_subnet_info(ex_gw_port, gw_p_sn['id'], is_primary=False)
            self._external_gateway_added(ri, ex_gw_port)

        list_port_ids_up.append(ex_gw_port['id'])

    def _process_gateway_cleared(self, ri, ex_gw_port):
        # We sort the port's subnets on subnet_id so we can be sure that
        # the same ip address is used as primary on the HA master router
        # as well as on all HA backup routers.
        gw_port_subnets = sorted(ex_gw_port['subnets'], key=itemgetter('id'))

        # Deconfigure the primary IP address
        self._set_subnet_info(ex_gw_port, gw_port_subnets[0]['id'])
        self._external_gateway_removed(ri, ex_gw_port)

        # Process the secondary subnets
        for gw_p_sn in gw_port_subnets[1:]:
            self._set_subnet_info(ex_gw_port, gw_p_sn['id'], is_primary=False)
            self._external_gateway_removed(ri, ex_gw_port)

    def _add_rid_to_vrf_list(self, ri):
        # not needed in base service helper
        pass

    def _remove_rid_from_vrf_list(self, ri):
        # not needed in base service helper
        pass

    def _process_router(self, ri):
        """Process a router, apply latest configuration and update router_info.

        Get the router dict from  RouterInfo and proceed to detect changes
        from the last known state. When new ports or deleted ports are
        detected, `internal_network_added()` or `internal_networks_removed()`
        are called accordingly. Similarly changes in ex_gw_port causes
         `external_gateway_added()` or `external_gateway_removed()` calls.
        Next, floating_ips and routes are processed. Also, latest state is
        stored in ri.internal_ports and ri.ex_gw_port for future comparisons.

        :param ri : RouterInfo object of the router being processed.
        :return:None
        :raises: networking_cisco.plugins.cisco.cfg_agent.cfg_exceptions.
        DriverException if the configuration operation fails.
        """
        try:
            ex_gw_port = ri.router.get('gw_port')
            ri.ha_info = ri.router.get('ha_info', None)
            gateway_set = ex_gw_port and not ri.ex_gw_port
            gateway_cleared = not ex_gw_port and ri.ex_gw_port
            internal_ports = ri.router.get(bc.constants.INTERFACE_KEY, [])
            # Once the gateway is set, then we know which VRF this router
            # belongs to. Keep track of it in our lists of routers, organized
            # as a dictionary by VRF name
            if gateway_set:
                self._add_rid_to_vrf_list(ri)

            new_ports, old_ports, change_details = (
                self._get_internal_port_changes(ri, internal_ports))

            list_port_ids_up = []

            non_global_router_roles = [None,
                                       c_constants.ROUTER_ROLE_HA_REDUNDANCY]
            if ri.router[ROUTER_ROLE_ATTR] in non_global_router_roles:
                self._process_new_ports(ri, new_ports, ex_gw_port,
                                        list_port_ids_up, change_details)
                self._process_old_ports(ri, old_ports, ex_gw_port,
                                        change_details)
            else:
                self._process_new_ports_global(ri, new_ports, ex_gw_port,
                                               list_port_ids_up)
                self._process_old_ports_global(ri, old_ports, ex_gw_port)

            if gateway_set:
                self._process_gateway_set(ri, ex_gw_port,
                                          list_port_ids_up)
            elif gateway_cleared:
                self._process_gateway_cleared(ri, ri.ex_gw_port)

            self._send_update_port_statuses(list_port_ids_up,
                                            bc.constants.PORT_STATUS_ACTIVE)
            if ex_gw_port:
                self._process_router_floating_ips(ri, ex_gw_port)

            global_router_roles = [c_constants.ROUTER_ROLE_GLOBAL,
                                   c_constants.ROUTER_ROLE_LOGICAL_GLOBAL]
            if ri.router[ROUTER_ROLE_ATTR] not in global_router_roles:
                self._enable_disable_ports(ri, ex_gw_port, internal_ports)

                if gateway_cleared:
                    # Remove this router from the list of routers by VRF
                    self._remove_rid_from_vrf_list(ri)

            ri.ex_gw_port = ex_gw_port
            self._routes_updated(ri)
        except cfg_exceptions.HAParamsMissingException as e:
            self._update_updated_routers_cache([ri.router_id])
            LOG.warning(e)
        except cfg_exceptions.DriverException as e:
            with excutils.save_and_reraise_exception():
                self._update_updated_routers_cache([ri.router_id])
                LOG.error(e)

    def _process_router_floating_ips(self, ri, ex_gw_port):
        """Process a router's floating ips.

        Compare floatingips configured in device (i.e., those fips in
        the ri.floating_ips "cache") with the router's updated floating ips
        (in ri.router.floating_ips) and determine floating_ips which were
        added or removed. Notify driver of the change via
        `floating_ip_added()` or `floating_ip_removed()`. Also update plugin
        with status of fips.

        :param ri:  RouterInfo object of the router being processed.
        :param ex_gw_port: Port dict of the external gateway port.
        :return: None
        :raises: networking_cisco.plugins.cisco.cfg_agent.cfg_exceptions.
        DriverException if the configuration operation fails.
        """

        # fips that exist in neutron db (i.e., the desired "truth")
        current_fips = ri.router.get(bc.constants.FLOATINGIP_KEY, [])
        # ids of fips that exist in neutron db
        current_fip_ids = {fip['id'] for fip in current_fips}
        # ids of fips that are configured in device
        configured_fip_ids = {fip['id'] for fip in ri.floating_ips}

        id_to_current_fip_map = {}

        fips_to_add = []
        # iterate of fips that exist in neutron db
        for configured_fip in current_fips:
            if configured_fip['port_id']:
                # store to later check if this fip has been remapped
                id_to_current_fip_map[configured_fip['id']] = configured_fip
                if configured_fip['id'] not in configured_fip_ids:
                    # Ensure that we add only after remove, in case same
                    # fixed_ip is mapped to different floating_ip within
                    # the same loop cycle. If add occurs before first,
                    # cfg will fail because of existing entry with
                    # identical fixed_ip
                    fips_to_add.append(configured_fip)

        fip_ids_to_remove = configured_fip_ids - current_fip_ids
        LOG.debug("fip_ids_to_add: %s" % fips_to_add)
        LOG.debug("fip_ids_to_remove: %s" % fip_ids_to_remove)

        fips_to_remove = []
        fip_statuses = {}
        # iterate over fips that are configured in device
        for configured_fip in ri.floating_ips:
            if configured_fip['id'] in fip_ids_to_remove:
                fips_to_remove.append(configured_fip)
                self._floating_ip_removed(
                    ri, ri.ex_gw_port, configured_fip['floating_ip_address'],
                    configured_fip['fixed_ip_address'])
                fip_statuses[configured_fip['id']] = (
                    bc.constants.FLOATINGIP_STATUS_DOWN)
                LOG.debug("Add to fip_statuses DOWN id:%s fl_ip:%s fx_ip:%s",
                          configured_fip['id'],
                          configured_fip['floating_ip_address'],
                          configured_fip['fixed_ip_address'])
            else:
                # handle possibly required remapping of a fip
                # ip address that fip currently is configured for
                configured_fixed_ip = configured_fip['fixed_ip_address']
                new_fip = id_to_current_fip_map[configured_fip['id']]
                # ip address that fip should be configured for
                current_fixed_ip = new_fip['fixed_ip_address']
                if (current_fixed_ip and configured_fixed_ip and
                        current_fixed_ip != configured_fixed_ip):
                    floating_ip = configured_fip['floating_ip_address']
                    self._floating_ip_removed(ri, ri.ex_gw_port,
                                              floating_ip, configured_fixed_ip)
                    fip_statuses[configured_fip['id']] = (
                        bc.constants.FLOATINGIP_STATUS_DOWN)
                    fips_to_remove.append(configured_fip)
                    fips_to_add.append(new_fip)

        for configured_fip in fips_to_remove:
            # remove fip from "cache" of fips configured in device
            ri.floating_ips.remove(configured_fip)

        for configured_fip in fips_to_add:
            self._floating_ip_added(ri, ex_gw_port,
                                    configured_fip['floating_ip_address'],
                                    configured_fip['fixed_ip_address'])
            # add fip to "cache" of fips configured in device
            ri.floating_ips.append(configured_fip)
            fip_statuses[configured_fip['id']] = (
                bc.constants.FLOATINGIP_STATUS_ACTIVE)
            LOG.debug("Add to fip_statuses ACTIVE id:%s fl_ip:%s fx_ip:%s",
                      configured_fip['id'],
                      configured_fip['floating_ip_address'],
                      configured_fip['fixed_ip_address'])

        if fip_statuses:
            LOG.debug("Sending floatingip_statuses_update: %s", fip_statuses)
            self.plugin_rpc.update_floatingip_statuses(
                self.context, ri.router_id, fip_statuses)

    def _router_added(self, router_id, router):
        """Operations when a router is added.

        Create a new RouterInfo object for this router and add it to the
        service helpers router_info dictionary.  Then `router_added()` is
        called on the device driver.

        :param router_id: id of the router
        :param router: router dict
        :return: None
        """
        ri = RouterInfo(router_id, router)
        driver = self.driver_manager.set_driver(router)
        if router[ROUTER_ROLE_ATTR] in [
            c_constants.ROUTER_ROLE_GLOBAL,
            c_constants.ROUTER_ROLE_LOGICAL_GLOBAL]:
            # No need to create a vrf for Global or logical global routers
            LOG.debug("Skipping router_added device processing for %(id)s as "
                      "its role is %(role)s",
                      {'id': router_id, 'role': router[ROUTER_ROLE_ATTR]})
        else:
            driver.router_added(ri)
        self.router_info[router_id] = ri

    def _router_removed(self, router_id, deconfigure=True):
        """Operations when a router is removed.

        Get the RouterInfo object corresponding to the router in the service
        helpers's router_info dict. If deconfigure is set to True,
        remove this router's configuration from the hosting device.
        :param router_id: id of the router
        :param deconfigure: if True, the router's configuration is deleted from
        the hosting device.
        :return: None
        """
        ri = self.router_info.get(router_id)
        if ri is None:
            LOG.warning("Info for router %s was not found. "
                        "Skipping router removal.", router_id)
            return
        ri.router['gw_port'] = None
        ri.router[bc.constants.INTERFACE_KEY] = []
        ri.router[bc.constants.FLOATINGIP_KEY] = []
        try:
            hd = ri.router['hosting_device']
            # We proceed to removing the configuration from the device
            # only if (a) deconfigure is set to True (default)
            # (b) the router's hosting device is reachable.
            if (deconfigure and
                    self._dev_status.is_hosting_device_reachable(hd)):
                self._process_router(ri)
                driver = self.driver_manager.get_driver(router_id)
                driver.router_removed(ri)
                self.driver_manager.remove_driver(router_id)
            del self.router_info[router_id]
            self._del_from_removed_routers_cache(router_id)
        except cfg_exceptions.DriverException:
            LOG.warning("Router remove for router_id: %s was incomplete. "
                        "Adding the router to removed_routers list",
                        router_id)
            self._update_removed_routers_cache([router_id])
            # remove this router from updated_routers if it is there. It might
            # end up there too if exception was thrown earlier inside
            # `_process_router()`
            self._del_from_updated_routers_cache(router_id)
        except ncc_errors.SessionCloseError as e:
            LOG.exception("ncclient Unexpected session close %s"
                          " while attempting to remove router", e)
            if not self._dev_status.is_hosting_device_reachable(hd):
                LOG.debug("Lost connectivity to Hosting Device %s" % hd['id'])
                # rely on heartbeat to detect HD state
                # and schedule resync when the device comes back
            else:
                # retry the router removal on the next pass
                self._update_removed_routers_cache([router_id])
                LOG.debug("Interim connectivity lost to hosting device %s, "
                          "enqueuing router %s in removed_routers set" %
                          pp.pformat(hd), router_id)

    def _internal_network_added(self, ri, port, ex_gw_port):
        driver = self.driver_manager.get_driver(ri.id)
        driver.internal_network_added(ri, port)
        if ri.snat_enabled and ex_gw_port:
            driver.enable_internal_network_NAT(ri, port, ex_gw_port)

    def _internal_network_removed(self, ri, port, ex_gw_port):
        driver = self.driver_manager.get_driver(ri.id)
        driver.internal_network_removed(ri, port)
        if ri.snat_enabled and ex_gw_port:
            #ToDo(Hareesh): Check if the intfc_deleted attribute is needed
            driver.disable_internal_network_NAT(ri, port, ex_gw_port,
                                                itfc_deleted=True)

    def _external_gateway_added(self, ri, ex_gw_port):
        driver = self.driver_manager.get_driver(ri.id)
        driver.external_gateway_added(ri, ex_gw_port)
        if ri.snat_enabled and ri.internal_ports:
            for port in ri.internal_ports:
                driver.enable_internal_network_NAT(ri, port, ex_gw_port)

    def _external_gateway_removed(self, ri, ex_gw_port):
        driver = self.driver_manager.get_driver(ri.id)
        if ri.snat_enabled and ri.internal_ports:
            for port in ri.internal_ports:
                driver.disable_internal_network_NAT(ri, port, ex_gw_port)
        driver.external_gateway_removed(ri, ex_gw_port)

    def _floating_ip_added(self, ri, ex_gw_port, floating_ip, fixed_ip):
        driver = self.driver_manager.get_driver(ri.id)
        driver.floating_ip_added(ri, ex_gw_port, floating_ip, fixed_ip)

    def _floating_ip_removed(self, ri, ex_gw_port, floating_ip, fixed_ip):
        driver = self.driver_manager.get_driver(ri.id)
        driver.floating_ip_removed(ri, ex_gw_port, floating_ip, fixed_ip)

    def _enable_router_interface(self, ri, port):
        driver = self.driver_manager.get_driver(ri.id)
        driver.enable_router_interface(ri, port)

    def _disable_router_interface(self, ri, port=None):
        driver = self.driver_manager.get_driver(ri.id)
        driver.disable_router_interface(ri, port)

    def _routes_updated(self, ri):
        """Update the state of routes in the router.

        Compares the current routes with the (configured) existing routes
        and detect what was removed or added. Then configure the
        logical router in the hosting device accordingly.
        :param ri: RouterInfo corresponding to the router.
        :return: None
        :raises: networking_cisco.plugins.cisco.cfg_agent.cfg_exceptions.
        DriverException if the configuration operation fails.
        """
        new_routes = ri.router['routes']
        old_routes = ri.routes
        adds, removes = bc.common_utils.diff_list_of_dict(old_routes,
                                                       new_routes)
        for route in adds:
            LOG.debug("Added route entry is '%s'", route)
            # remove replaced route from deleted route
            for del_route in removes:
                if route['destination'] == del_route['destination']:
                    removes.remove(del_route)
            driver = self.driver_manager.get_driver(ri.id)
            driver.routes_updated(ri, 'replace', route)

        for route in removes:
            LOG.debug("Removed route entry is '%s'", route)
            driver = self.driver_manager.get_driver(ri.id)
            driver.routes_updated(ri, 'delete', route)
        ri.routes = new_routes

    @staticmethod
    def _set_subnet_info(port, subnet_id, is_primary=True):
        ip = next((i['ip_address'] for i in port['fixed_ips']
                   if i['subnet_id'] == subnet_id), None)
        if ip is None:
            raise IPAddressMissingException(port_id=port['id'],
                                            subnet_id=subnet_id)
        subnet = next(sn for sn in port['subnets'] if sn['id'] == subnet_id)
        prefixlen = netaddr.IPNetwork(subnet['cidr']).prefixlen
        port['ip_info'] = {'subnet_id': subnet_id, 'is_primary': is_primary,
                           'ip_cidr': "%s/%s" % (ip, prefixlen)}



# Copyright 2017 Cisco Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from networking_cisco.ml2_drivers.ucsm import constants as const

from oslo_config import cfg

from tempest.api.network import base
from tempest import config
from tempest.lib import decorators

from ucsmsdk.ucshandle import UcsHandle

CONF = config.CONF

tempest_ucsm_opts = [
    cfg.StrOpt('service_profile'),
    cfg.StrOpt('ucsm_ip'),
    cfg.StrOpt('ucsm_username'),
    cfg.StrOpt('ucsm_password'),
    cfg.ListOpt('ucsm_virtio_eth_ports', default=[const.ETH0, const.ETH1])
]

CONF.register_opts(tempest_ucsm_opts, 'ucsm')


class TestUCSMConfiguration(base.BaseAdminNetworkTest):

    """This tests UCSM configuration with ml2 driver

    Steps:
    1. Create network
    2. Create subnet
    3. Query UCSM for changes
    4. Check VLAN set on UCSM

    """
    @decorators.attr(type='smoke')
    @decorators.idempotent_id('bdbb5441-9204-419d-a225-b4fdbfb1a1a8')
    def test_ucsm_config_scenario(self):
        # Create a network
        network = self.create_network()
        self.addCleanup(self.networks_client.delete_network, network['id'])
        self.assertEqual('ACTIVE', network['status'])
        # Find a cidr that is not in use yet and create a subnet with it
        self.create_subnet(network)

        admin_network = self.admin_networks_client.show_network(network['id'])
        network_vlan = admin_network['network']['provider:segmentation_id']

        # Connect to UCSM
        handle = UcsHandle(CONF.ucsm.ucsm_ip, CONF.ucsm.ucsm_username,
                CONF.ucsm.ucsm_password)
        handle.login()

        # Check Service profile is on UCSM
        sp = handle.query_dn('org-root/ls-' + CONF.ucsm.service_profile)
        self.assertIsNotNone(sp)

        # Sleep while waiting for neutron to create ports
        time.sleep(10)

        # Check eth ports are in sevice profile
        for interface in CONF.ucsm.ucsm_virtio_eth_ports:
            eth = handle.query_dn('org-root/ls-' +
                                  CONF.ucsm.service_profile +
                                  '/ether-' +
                                  interface)
            self.assertIsNotNone(eth)

        # Check vlan exists
        vlan = handle.query_dn('fabric/lan/net-OS-' + str(network_vlan))
        self.assertEqual(str(network_vlan), getattr(vlan, 'id', None))

        # Check vlan is associated with a vnic on service profile
        for interface in CONF.ucsm.ucsm_virtio_eth_ports:
            vnic_vlan_eth = handle.query_dn('org-root/ls-' +
                                            CONF.ucsm.service_profile +
                                            '/ether-' +
                                            interface +
                                            '/if-OS-' +
                                            str(network_vlan))
            self.assertEqual(str(network_vlan),
                             getattr(vnic_vlan_eth, 'vnet', None))



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


DFA_db_session = None

Base = declarative_base()


def configure_db(cfg):
    global DFA_db_session

    if DFA_db_session:
        return

    connection = cfg.dfa_mysql.connection
    engine = create_engine(connection, echo=False)
    Base.metadata.create_all(engine)
    Session = sessionmaker(bind=engine, autocommit=True)
    DFA_db_session = Session()


def get_session():
    return DFA_db_session



# Copyright 2017 Cisco Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add VPC alloc table

Revision ID: f3765e42de23
Revises: 9148d96f9b39
Create Date: 2017-06-07 17:10:09.597016

"""

# revision identifiers, used by Alembic.
revision = 'f3765e42de23'
down_revision = '9148d96f9b39'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('cisco_ml2_nexus_vpc_alloc',
        sa.Column('switch_ip', sa.String(length=64), nullable=False,
                  index=True),
        sa.Column('vpc_id', sa.Integer(), nullable=False),
        sa.Column('learned', sa.Boolean(), nullable=False),
        sa.Column('active', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('switch_ip', 'vpc_id')
    )
